from typing import List

import pandas as pd

from .pandas_to_welly import WellyToSubsurfaceHelper
from ._well_files_reader import read_borehole_files, read_collar
from ._welly_reader import welly_to_subsurface

from subsurface.reader.readers_data import ReaderWellsHelper, ReaderFilesHelper
from subsurface.structs import UnstructuredData


def read_wells_to_unstruct(reader_wells_helper: ReaderWellsHelper,
                           backend='welly', n_vertex_per_well=80,
                           table: List['welly.Component'] = None) -> UnstructuredData:
    """Read from csv files (or excel) to `subsurface.Unstructured` object.

    Args:
        backend (string): Which library use for reading and processing of data.
         So far: welly
        table (List[Striplog.Component]): List of components to map lithologies
         to value.
        n_vertex_per_well (int): Number of vertex used to describe the geometry of the
         well.

    Returns:
        `subsurface.UnstructuredData`:  if `return_welly` also the
         welly object

    """
    pandas_dict = read_borehole_files(reader_wells_helper)

    if backend == 'welly':
        wts = WellyToSubsurfaceHelper(**pandas_dict)
        unstruct = welly_to_subsurface(wts, n_vertex_per_well=n_vertex_per_well, table=table)
    else:
        raise AttributeError('Only welly is available at the moment')

    return unstruct


def borehole_location_to_unstruct(reader_helper: ReaderFilesHelper,
                                  add_number_segments: bool = True) -> UnstructuredData:
    collars = read_collar(reader_helper)
    collars_attributes = pd.DataFrame()

    # Remove duplicates
    collars_single_well = collars[~collars.index.duplicated()]
    wells_names = collars_single_well.index

    if add_number_segments is True:
        number_of_segments = collars.index.value_counts(sort=False).values
        collars_attributes['number_segments'] = number_of_segments

    ud = UnstructuredData.from_array(
        vertex=collars_single_well[['x', 'y', 'altitude']].values.astype('float32'),
        cells="points",
        cells_attr=collars_attributes.astype('float32'),
        xarray_attributes={"wells_names": wells_names.values.tolist()})  # TODO: This should be int16!

    return ud



from dataclasses import dataclass, field

import numpy as np


@dataclass
class GOCADMesh:
    header: dict = field(default_factory=dict)
    coordinate_system: dict = field(default_factory=dict)
    property_class_headers: list = field(default_factory=list)
    vertices: np.ndarray = field(default_factory=lambda: np.empty((0, 3)))
    vertex_indices: np.ndarray = field(default_factory=lambda: np.array([]))
    edges: np.ndarray = field(default_factory=lambda: np.empty((0, 3), dtype=int))
    bstones: list = field(default_factory=list)
    borders: list = field(default_factory=list)
    metadata: dict = field(default_factory=dict)

    @property
    def vectorized_edges(self):
        # Create index mapping from original to zero-based indices
        idx_map = {old_idx: new_idx for new_idx, old_idx in enumerate(self.vertex_indices)}
        # Map triangle indices
        try:
            triangles_mapped = np.vectorize(idx_map.get)(self.edges)
        except TypeError as e:
            self._verbose_debugging()
            raise f"Error mapping indices for mesh: {e}"

        return triangles_mapped

    @property
    def color(self):
        """Try to get the color from the metadata. Can be in the form of:
        *solid*color: #87ceeb or *solid*color: 0 0 1 1 (rgba).
        Returns a color in the format acceptable by PyVista.
        """
        color = None

        # First try to find a color value from the header
        for key, value in self.header.items():
            if 'color' in key.lower():
                color = value.strip()
                break

        # If no color was found, return None
        if not color:
            return None

        # Handle hexadecimal color string (e.g., #87ceeb)
        if color.startswith('#') and len(color) == 7:
            return color  # already valid as a hex string

        # Handle space-separated RGBA or RGB values (e.g., "0 0 1 1")
        if ' ' in color:
            color_vals = [float(c) for c in color.split()]
            if len(color_vals) == 4:  # RGBA
                return color_vals[:3]  # ignore the alpha channel for now, as PyVista handles RGB
            elif len(color_vals) == 3:  # RGB
                return color_vals  # already in the right format

        # Fallback: if none of the formats match, return None
        return None

    def _verbose_debugging(self):
        # Create index mapping from original to zero-based indices
        idx_map = {old_idx: new_idx for new_idx, old_idx in enumerate(self.vertex_indices)}

        # Check for missing indices
        unique_edge_indices = np.unique(self.edges)
        missing_indices = set(unique_edge_indices) - set(idx_map.keys())
        if missing_indices:
            raise ValueError(f"Edges contain indices not found in vertex_indices: {missing_indices}")

        # Map triangle indices using a list comprehension
        try:
            edges_flat = self.edges.flatten()
            mapped_flat = [idx_map[idx] for idx in edges_flat]
            triangles_mapped = np.array(mapped_flat).reshape(self.edges.shape)
        except Exception as e:
            raise Exception(f"Error mapping indices for mesh: {e}")

        return triangles_mapped



from typing import Union

from subsurface import optional_requirements
from ....core.structs.base_structures import StructuredData
import numpy as np


def read_in_segy(filepath: str, ignore_geometry: bool = True, flip_y_axis: bool = True) -> StructuredData:
    """
    Reads a SEG-Y file and processes its data into a structured format.

    This function opens a SEG-Y file using the `segyio` library and extracts 
    the trace data. The data is optionally flipped along the y-axis and its 
    axes are rearranged before being converted into a `StructuredData` object. 
    The `ignore_geometry` option determines whether SEG-Y header geometry 
    information is considered during file read.

    Args:
        filepath (str): The path to the SEG-Y file to be read.
        ignore_geometry (bool): Whether to ignore the geometry information 
            from the SEG-Y header. Defaults to True.
        flip_y_axis (bool): Whether to flip the data vertically (on the y-axis). 
            Defaults to True.

    Returns:
        StructuredData: An instance of `StructuredData` containing the 
        processed trace data from the SEG-Y file.

    Raises:
        IOError: If there is an error in opening or reading the SEG-Y file.
        ValueError: If the data cannot be processed into the expected format.
    """
    segyio = optional_requirements.require_segyio()
    segyfile = segyio.open(filepath, ignore_geometry=ignore_geometry)

    data = np.asarray([np.copy(tr) for tr in segyfile.trace[:]])
    
    if flip_y_axis:
        data = np.flip(data, axis=1)
        
    data = np.swapaxes(data, 0, 1)
    
    sd = StructuredData.from_numpy(data)  # data holds traces * (samples per trace) values
    segyfile.close()
    return sd


def create_mesh_from_coords(coords: dict, zmin: Union[float, int], zmax: Union[float, int] = 0.0):
    """Creates a mesh for plotting StructuredData

    Args:
        coords (Union[dict, LineString]): the x and y, i.e. latitude and longitude, location of the traces of the seismic profile
        zmax (float): the maximum elevation of the seismic profile, by default 0.0
        zmin (float): the location in z where the lowest sample was taken

    Returns: vertices and faces for creating an UnstructuredData object

    """
    n = len(coords['x'])
    coords = np.array([coords['x'], coords['y']]).T
    # duplicating the line, once with z=lower and another with z=upper values
    vertices = np.zeros((2 * n, 3))
    vertices[:n, :2] = coords
    vertices[:n, 2] = zmin
    vertices[n:, :2] = coords
    vertices[n:, 2] = zmax
    # i+n --- i+n+1
    # |\      |
    # | \     |
    # |  \    |
    # |   \   |
    # i  --- i+1

    scipy = optional_requirements.require_scipy()
    tri = scipy.spatial.qhull.Delaunay(vertices[:, [0, 2]])
    faces = tri.simplices
    return vertices, faces


def apply_colormap_to_texture(texture: StructuredData, cmap_name="bwr"):
    """
    Convert a single-channel seismic texture.data array into RGB
    using a Matplotlib colormap (e.g., 'bwr', 'jet', 'RdBu_r', etc.).
    """
    # 'texture.data' should be a 2D array of amplitudes: shape = (height, width)
    import matplotlib.colors as colors
    import matplotlib.pyplot as plt
    data = texture.values
    # 1. Normalize data to [0,1] range based on its min/max
    min_val, max_val = data.min(), data.max()
    norm = colors.Normalize(vmin=min_val, vmax=max_val)

    norm = colors.Normalize(vmin=-6, vmax=6)

    # 2. Get a Matplotlib colormap
    cmap = plt.get_cmap(cmap_name)

    # 3. Map normalized data -> RGBA, shape becomes (height, width, 4)
    rgba_data = cmap(norm(data))

    # 4. Convert from float in [0,1] to uint8 in [0,255], and drop alpha channel
    rgb_data = (rgba_data[..., :3] * 255).astype(np.uint8)

    texture = StructuredData.from_numpy(rgb_data)
    return texture



import enum
from typing import Union, TextIO, Optional
import io
import os

import numpy as np
from ....core.structs import UnstructuredData
from .... import optional_requirements
from ....core.structs import TriSurf, StructuredData


class TriMeshTransformations(enum.Flag):
    UP_Z = 2**1
    UP_Y = 2**2
    FORWARD_MINUS_Z = 2**3
    FORWARD_PLUS_Z = 2**4
    RIGHT_HANDED_Z_UP_Y_REVERSED = UP_Y | FORWARD_MINUS_Z
    RIGHT_HANDED_Z_UP = UP_Y | FORWARD_PLUS_Z
    ORIGINAL = UP_Z | FORWARD_MINUS_Z


def load_with_trimesh(path_to_file_or_buffer, file_type: Optional[str] = None,
                      coordinate_system: TriMeshTransformations = TriMeshTransformations.RIGHT_HANDED_Z_UP, *, plot=False):
    """
    Load a mesh with trimesh and convert to the specified coordinate system.

    """
    trimesh = optional_requirements.require_trimesh()
    scene_or_mesh = LoadWithTrimesh.load_with_trimesh(path_to_file_or_buffer, file_type, plot)

    match coordinate_system:
        case TriMeshTransformations.ORIGINAL:
            return scene_or_mesh
        # * Forward -Z up Y
        case TriMeshTransformations.RIGHT_HANDED_Z_UP:
            # Transform from Y-up (modeling software) to Z-up (scientific)
            # This rotates the model so that:
            # Old Y axis → New Z axis (pointing up)
            # Old Z axis → New -Y axis
            # Old X axis → Remains as X axis
            transform = np.array([
                    [1, 0, 0, 0],  
                    [0, 0, -1, 0],  
                    [0, 1, 0, 0], 
                    [0, 0, 0, 1]
            ])
        case TriMeshTransformations.RIGHT_HANDED_Z_UP_Y_REVERSED:
            # * Forward Z Up Y
            transform=np.array([
                    [-1, 0, 0, 0],
                    [0, 0, 1, 0],
                    [0, 1, 0, 0],
                    [0, 0, 0, 1],
            ])
            # Apply the coordinate transformation
        # TODO: Add all the options of blender
        case _:
            raise ValueError(f"Invalid coordinate system: {coordinate_system}")

    if isinstance(scene_or_mesh, trimesh.Scene):
        for geometry in scene_or_mesh.geometry.values():
            geometry.apply_transform(transform)
    else:
        scene_or_mesh.apply_transform(transform)
    return scene_or_mesh


def trimesh_to_unstruct(scene_or_mesh: Union["trimesh.Trimesh", "trimesh.Scene"]) -> TriSurf:
    return TrimeshToSubsurface.trimesh_to_unstruct(scene_or_mesh)


class LoadWithTrimesh:
    @classmethod
    def load_with_trimesh(cls, path_to_file_or_buffer, file_type: Optional[str] = None, plot=False):
        trimesh = optional_requirements.require_trimesh()
        # Load the OBJ with Trimesh using the specified options
        scene_or_mesh = trimesh.load(
            file_obj=path_to_file_or_buffer,
            file_type=file_type,
            force="mesh"
        )
        # Process single mesh vs. scene
        if isinstance(scene_or_mesh, trimesh.Scene):
            print("Loaded a Scene with multiple geometries.")
            cls._process_scene(scene_or_mesh)
            if plot:
                scene_or_mesh.show()
        else:
            print("Loaded a single Trimesh object.")
            print(f" - Vertices: {len(scene_or_mesh.vertices)}")
            print(f" - Faces: {len(scene_or_mesh.faces)}")
            cls.handle_material_info(scene_or_mesh)
            if plot:
                scene_or_mesh.show()

        return scene_or_mesh

    @classmethod
    def handle_material_info(cls, geometry):
        """
        Handle and print material information for a single geometry,
        explicitly injecting the PIL image if provided.
        """
        if geometry.visual and hasattr(geometry.visual, 'material'):
            material = geometry.visual.material

            print("Trimesh material:", material)

            # If there's already an image reference in the material, let the user know
            if hasattr(material, 'image') and material.image is not None:
                print("  -> Material already has an image:", material.image)
            
            if geometry.visual.uv is None:
                raise ValueError("Geometry does not have UV coordinates for texture mapping, despite having a material."
                                 "This can also happen if the geometry is given in quads instead of triangles.")
        else:
            print("No material found or no 'material' attribute on this geometry.")

    @classmethod
    def _process_scene(cls, scene):
        """Process a scene with multiple geometries."""
        geometries = scene.geometry
        assert len(geometries) > 0, "No geometries found in the scene."

        print(f"Loaded a Scene with {len(scene.geometry)} geometry object(s).")
        for geom_name, geom in geometries.items():
            print(f" Submesh: {geom_name}")
            print(f"  - Vertices: {len(geom.vertices)}")
            print(f"  - Faces: {len(geom.faces)}")

            print(f"Geometry '{geom_name}':")
            cls.handle_material_info(geom)


class TrimeshToSubsurface:
    @classmethod
    def trimesh_to_unstruct(cls, scene_or_mesh: Union["trimesh.Trimesh", "trimesh.Scene"]) -> TriSurf:
        """
        Convert a Trimesh or Scene object to a subsurface TriSurf object.

        This function takes either a `trimesh.Trimesh` object or a `trimesh.Scene` 
        object and converts it to a `subsurface.TriSurf` object. If the input is 
        a scene containing multiple geometries, it processes all geometries and 
        combines them into a single TriSurf object. If the input is a single 
        Trimesh object, it directly converts it to a TriSurf object. An error 
        is raised if the input is neither a `trimesh.Trimesh` nor a `trimesh.Scene` 
        object.

        Parameters:
            scene_or_mesh (Union[trimesh.Trimesh, trimesh.Scene]): 
                Input geometry data, either as a Trimesh object representing 
                a single mesh or a Scene object containing multiple geometries.

        Note:
            ! Multimesh with multiple materials will read the uvs but not the textures since in that case is better
            ! to read directly the multiple images (compressed) whenever the user wants to work with them. 

        Returns:
            subsurface.TriSurf: Converted subsurface representation of the 
            provided geometry data.

        Raises:
            ValueError: If the input is neither a `trimesh.Trimesh` object nor 
            a `trimesh.Scene` object.
        """
        trimesh = optional_requirements.require_trimesh()
        if isinstance(scene_or_mesh, trimesh.Scene):
            # Process scene with multiple geometries
            ts = cls._trisurf_from_scene(scene_or_mesh, trimesh)

        elif isinstance(scene_or_mesh, trimesh.Trimesh):
            ts = cls._trisurf_from_trimesh(scene_or_mesh)


        else:
            raise ValueError("Input must be a Trimesh object or a Scene with multiple geometries.")

        return ts

    @classmethod
    def _trisurf_from_trimesh(cls, scene_or_mesh):
        # Process single mesh
        tri = scene_or_mesh
        pandas = optional_requirements.require_pandas()
        frame = pandas.DataFrame(tri.face_attributes)
        # Check frame has a valid shape for cells_attr if not make None
        if frame.shape[0] != tri.faces.shape[0]:
            frame = None
        # Get UV coordinates if they exist
        vertex_attr = None
        if hasattr(tri.visual, 'uv') and tri.visual.uv is not None:
            vertex_attr = pandas.DataFrame(
                tri.visual.uv,
                columns=['u', 'v']
            )
        unstruct = UnstructuredData.from_array(
            np.array(tri.vertices),
            np.array(tri.faces),
            cells_attr=frame,
            vertex_attr=vertex_attr,
            xarray_attributes={
                    "bounds": tri.bounds.tolist(),
            },
        )

        texture = cls._extract_texture_from_material(tri)

        ts = TriSurf(
            mesh=unstruct,
            texture=texture,
        )
        return ts

    @classmethod
    def _trisurf_from_scene(cls, scene_or_mesh: 'Scene', trimesh: 'trimesh') -> TriSurf:
        pandas = optional_requirements.require_pandas()
        geometries = scene_or_mesh.geometry
        assert len(geometries) > 0, "No geometries found in the scene."
        all_vertex = []
        all_cells = []
        cell_attr = []
        all_vertex_attr = []
        _last_cell = 0
        texture = None
        for i, (geom_name, geom) in enumerate(geometries.items()):
            geom: trimesh.Trimesh
            LoadWithTrimesh.handle_material_info(geom)

            # Append vertices
            all_vertex.append(np.array(geom.vertices))

            # Adjust cell indices and append
            cells = np.array(geom.faces)
            if len(all_cells) > 0:
                cells = cells + _last_cell
            all_cells.append(cells)

            # Create attribute array for this geometry
            cell_attr.append(np.ones(len(cells)) * i)

            _last_cell = cells.max() + 1

            # Get UV coordinates if they exist
            if hasattr(geom.visual, 'uv') and geom.visual.uv is not None:
                vertex_attr = pandas.DataFrame(
                    geom.visual.uv,
                    columns=['u', 'v']
                )
                all_vertex_attr.append(vertex_attr)

            # Extract texture from material if it is only one geometry
            if len(geometries) == 1:
                texture = cls._extract_texture_from_material(geom)

        # Create the combined UnstructuredData
        unstruct = UnstructuredData.from_array(
            vertex=np.vstack(all_vertex),
            cells=np.vstack(all_cells),
            vertex_attr=pandas.concat(all_vertex_attr, ignore_index=True) if len(all_vertex_attr) > 0 else None,
            cells_attr=pandas.DataFrame(np.hstack(cell_attr), columns=["Geometry id"]),
            xarray_attributes={
                    "bounds": scene_or_mesh.bounds.tolist(),
            },
        )

        # If there is a texture
        ts = TriSurf(
            mesh=unstruct,
            texture=texture,
        )

        return ts

    @classmethod
    def _extract_texture_from_material(cls, geom):
        from PIL.JpegImagePlugin import JpegImageFile
        from PIL.PngImagePlugin import PngImageFile
        import trimesh

        if geom.visual is None or getattr(geom.visual, 'material', None) is None:
            return None

        array = np.empty(0)
        if isinstance(geom.visual.material, trimesh.visual.material.SimpleMaterial):
            image: JpegImageFile = geom.visual.material.image
            if image is None:
                return None
            array = np.array(image)
        elif isinstance(geom.visual.material, trimesh.visual.material.PBRMaterial):
            image: PngImageFile = geom.visual.material.baseColorTexture
            array = np.array(image.convert('RGBA'))

            if image is None:
                return None
        else:
            raise ValueError(f"Unsupported material type: {type(geom.visual.material)}")

        # Asser that image has 3 channels    assert array.shape[2] == 3    from PIL.PngImagePlugin import PngImageFile
        assert array.shape[2] == 3 or array.shape[2] == 4
        texture = StructuredData.from_numpy(array)
        return texture

    @classmethod
    def _validate_texture_path(cls, texture_path):
        """Validate the texture file path."""
        if texture_path and not texture_path.lower().endswith(('.png', '.jpg', '.jpeg')):
            raise ValueError("Texture path must be a PNG or JPEG file")


class TriMeshReaderFromBlob:
    @classmethod
    def OBJ_stream_to_trisurf(cls, obj_stream: TextIO, mtl_stream: list[TextIO],
                              texture_stream: list[io.BytesIO], coord_system: TriMeshTransformations) -> TriSurf:
        """
        Load an OBJ file from a stream and convert it to a TriSurf object.
        
        Parameters:
            obj_stream: TextIO containing the OBJ file data (text format)
            mtl_stream: TextIO containing the MTL file data (text format)
            texture_stream: BytesIO containing the texture file data (binary format)
        
        Returns:
            TriSurf: The loaded mesh with textures if available
        """
        trimesh = optional_requirements.require_trimesh()
        import tempfile

        path_in = "file.obj"

        # Create a temporary directory to store associated files
        with tempfile.TemporaryDirectory() as temp_dir:
            # Write the OBJ content to a temp file
            obj_path = os.path.join(temp_dir, os.path.basename(path_in))
            with open(obj_path, 'w') as f:  # Use text mode 'w' for text files
                obj_stream.seek(0)
                f.write(obj_stream.read())
                obj_stream.seek(0)

            if mtl_stream is not None:
                cls.write_material_files(
                    mtl_streams=mtl_stream,
                    obj_stream=obj_stream,
                    temp_dir=temp_dir,
                    texture_streams=texture_stream
                )

            # Now load the OBJ with all associated files available
            scene_or_mesh = load_with_trimesh(
                path_to_file_or_buffer=obj_path,
                file_type="obj",
                coordinate_system=coord_system
            )

            # Convert to a TriSurf object
            tri_surf = TrimeshToSubsurface.trimesh_to_unstruct(scene_or_mesh)

        return tri_surf

    @classmethod
    def write_material_files(cls, mtl_streams: list[TextIO], obj_stream: TextIO, temp_dir, texture_streams: list[io.BytesIO]):
        # Extract mtl references from the OBJ file
        mtl_files = cls._extract_mtl_references(obj_stream)
        # Download and save MTL files
        for e, mtl_file in enumerate(mtl_files):
            mtl_path = f"{temp_dir}/{mtl_file}" if temp_dir else mtl_file
            mtl_stream = mtl_streams[e] if mtl_streams else None
            try:
                # Save the MTL file to temp directory
                mtl_temp_path = os.path.join(temp_dir, mtl_file)
                with open(mtl_temp_path, 'w') as f:  # Use text mode 'w' for text files
                    mtl_stream.seek(0)
                    f.write(mtl_stream.read())

                # Extract texture references from MTL
                mtl_stream.seek(0)
                texture_files = cls._extract_texture_references(mtl_stream)

                if texture_streams is None:
                    continue

                # Download texture files
                for ee, texture_file in enumerate(texture_files):
                    texture_path = f"{temp_dir}/{texture_file}" if temp_dir else texture_file
                    texture_stream = texture_streams[ee] if texture_streams else None
                    try:
                        # Save the texture file to temp directory
                        with open(os.path.join(temp_dir, texture_file), 'wb') as f:  # Binary mode for textures
                            texture_stream.seek(0)
                            f.write(texture_stream.read())
                    except Exception as e:
                        print(f"Failed to load texture {texture_file}: {e}")
            except Exception as e:
                print(f"Failed to load MTL file {mtl_file}: {e}")

    @classmethod
    def _extract_mtl_references(cls, obj_stream):
        """Extract MTL file references from an OBJ file."""
        obj_stream.seek(0)
        mtl_files = []

        # TextIO stream already contains decoded text, so no need to decode
        obj_text = obj_stream.read()
        obj_stream.seek(0)

        for line in obj_text.splitlines():
            if line.startswith('mtllib '):
                mtl_name = line.split(None, 1)[1].strip()
                mtl_files.append(mtl_name)

        return mtl_files

    @classmethod
    def _extract_texture_references(cls, mtl_stream):
        """
        Extract texture file references from an MTL file.
        Works with both TextIO and BytesIO streams.
        
        Parameters:
            mtl_stream: TextIO or BytesIO containing the MTL file data
            
        Returns:
            list[str]: List of texture file names referenced in the MTL
        """
        mtl_stream.seek(0)
        texture_files = []

        # Handle both TextIO and BytesIO
        if isinstance(mtl_stream, io.TextIOWrapper):
            # TextIO stream already contains decoded text
            mtl_text = mtl_stream.read()
        else:
            # BytesIO stream needs to be decoded
            mtl_text = mtl_stream.read().decode('utf-8', errors='replace')

        mtl_stream.seek(0)

        for line in mtl_text.splitlines():
            # Check for texture map definitions
            for prefix in ['map_Kd ', 'map_Ka ', 'map_Ks ', 'map_Bump ', 'map_d ']:
                if line.startswith(prefix):
                    parts = line.split(None, 1)
                    if len(parts) > 1:
                        texture_name = parts[1].strip()
                        texture_files.append(texture_name)
                    break

        return texture_files



from subsurface.writer.to_rex.common import mesh_header_size, \
    write_data_block_header, encode
from subsurface.writer.to_rex.data_struct import RexMesh


__all__ = ['mesh_encode', 'write_mesh_coordinates', 'write_mesh_header']


def mesh_encode(rex_mesh: RexMesh, data_id: int):
    material_id = rex_mesh.material_id
    n_vtx_coord = rex_mesh.n_vtx
    n_triangles = rex_mesh.n_triangles
    n_vtx_colors = rex_mesh.n_color
    surface_name = rex_mesh.name
    ver_ravel = rex_mesh.ver_ravel
    tri_ravel = rex_mesh.tri_ravel
    c_r = rex_mesh.color_ravel

    # Write Mesh block - header
    mesh_header_bytes = write_mesh_header(
        n_vtx_coord / 3, n_triangles / 3,
        n_vtx_colors=n_vtx_colors / 3,
        start_vtx_coord=mesh_header_size,
        start_nor_coord=mesh_header_size + n_vtx_coord * 4,
        start_tex_coord=mesh_header_size + n_vtx_coord * 4,
        start_vtx_colors=mesh_header_size + n_vtx_coord * 4,
        start_triangles=mesh_header_size +
                        ((n_vtx_coord + n_vtx_colors) * 4),
        name=surface_name,
        material_id=material_id  # self.data_id + surface_df.shape[0]
    )

    # Write Mesh block - Vertex, triangles
    mesh_block_bytes = write_mesh_coordinates(ver_ravel,
                                              tri_ravel,
                                              colors=c_r  # When using
                                              # material we can avoid this
                                              )

    # Calculate the size of the mesh block
    mesh_block_size_no_data_block_header = len(mesh_header_bytes) + \
                                           len(mesh_block_bytes)  # This is cte 128

    # Write data block header for Mesh 1
    data_header_bytes = write_data_block_header(
        size_data=mesh_block_size_no_data_block_header,
        data_id=data_id,
        data_type=3,  # 3 for mesh
        version_data=1  # Probably useful for counting
        # the operation number
    )

    rex_bytes = data_header_bytes + mesh_header_bytes + mesh_block_bytes
    return rex_bytes


def write_mesh_coordinates(vertex, triangles, normal=None, texture=None,
                           colors=None):
    """Block with the coordinates of a mesh. This has to go with a header!

    Args:
        vertex (numpy.ndarray[float32]): Array of vertex XYZXYZ...
        triangles (numpy.ndarray[int32]): This is a list of integers which form
         one triangle. Please make sure that normal and texture coordinates are inline with the
         vertex coordinates. One index refers to the same normal and texture position. The
         triangle orientation is required to be counter-clockwise (CCW)
        normal (numpy.ndarray):
        texture (numpy.ndarray):
        colors (numpy.ndarray):

    Returns:

    """

    # ver = vertex.ravel()
    # tri = triangles.ravel()
    if normal is None:
        normal = []
    if texture is None:
        texture = []
    if colors is None:
        colors = []

    input_ = [(vertex, 'float32'),
              (normal, 'float32'),
              (texture, 'float32'),
              (colors, 'float32'),
              (triangles, 'uint32')]

    block_bytes = encode(input_)
    return block_bytes


def write_mesh_header(n_vtx_coord, n_triangles,
                      start_vtx_coord, start_nor_coord, start_tex_coord,
                      start_vtx_colors,
                      start_triangles,
                      name, material_id=1,  # material_id=9223372036854775807
                      n_nor_coord=0, n_tex_coord=0, n_vtx_colors=0,
                      lod=1, max_lod=1):
    """Function to write MESH DATA BLOCK header. The header size is fixed at 128 bytes.

    Args:
        n_vtx_coord: number of vertex coordinates
        n_triangles: number of triangles
        start_vtx_coord: start vertex coordinate block (relative to mesh block start)
        start_nor_coord: start vertex normals block (relative to mesh block start)
        start_tex_coord: start of texture coordinate block (relative to mesh block start)
        start_vtx_colors: start of colors block (relative to mesh block start)
        start_triangles: start triangle block for vertices (relative to mesh block start)
        name (str): Name of the mesh
        material_id (int):  id which refers to the corresponding material block in this file
        n_nor_coord:  number of normal coordinates (can be zero)
        n_tex_coord:  number of texture coordinates (can be zero)
        n_vtx_colors: number of vertex colors (can be zero)
        lod (int): level of detail for the given geometry
        max_lod (int): maximal level of detail for given geometry

    Returns:
        bytes: array of bytes
    """

    # Strings are immutable so there is no way to modify them in place
    str_size = len(name)  # Size of the actual name of the mesh
    rest_name = ' ' * (74 - str_size)  #
    full_name = name + rest_name

    input_ = [([lod, max_lod], 'uint16'),  # Level of detail
              ([n_vtx_coord,  # number of vertex coordinates
                n_nor_coord,  # number of normal coordinates (can be zero)
                n_tex_coord,  # number of texture coordinates (can be zero)
                n_vtx_colors,  # number of vertex colors (can be zero)
                n_triangles,  # number of triangles
                start_vtx_coord,
                # start vertex coordinate block (relative to mesh block start)
                start_nor_coord,
                # start vertex normals block (relative to mesh block start)
                start_tex_coord,
                # start of texture coordinate block (relative to mesh block start)
                start_vtx_colors,
                # start of colors block (relative to mesh block start)
                start_triangles
                # start triangle block for vertices (relative to mesh block start)
                ],
               'uint32'),
              (material_id, 'uint64'),
              # id which refers to the corresponding material block in this file
              (str_size, 'uint16'),  # size of the following string name
              (full_name, 'bytes')]  # name of the mesh (this is user-readable)

    block_bytes = encode(input_)
    return block_bytes



import warnings

import xarray as xr

from subsurface import UnstructuredData, StructuredData


__all__ = ['read_unstruct', 'read_struct']


def read_unstruct(path, legacy=False, **kwargs):
    ds = xr.open_dataset(path, **kwargs)

    if legacy is False:
        try:
            ud = UnstructuredData(ds)
        except KeyError:
            warnings.warn("Trying loading legacy files.")
            ud = read_unstruct(path, legacy=True)
    else:
        ud = _legacy_read_unstruct(ds)
    return ud


def _legacy_read_unstruct(ds):
    xarray_dict = {
        "vertex": ds.vertex,
        "cells": ds.cells,
        "cell_attrs": ds.attributes.rename({"attribute": "cell_attr"}),
        "vertex_attrs": ds.points_attributes.rename({"points_attribute": "vertex_attr"})
    }
    ud = UnstructuredData.from_data_arrays_dict(xarray_dict, ds.coords, ds.attrs)
    return ud


def read_struct(path, **kwargs):
    ds = xr.open_dataset(path, **kwargs)
    return StructuredData(ds)



from .csv_mesh_reader import mesh_csv_to_vertex, mesh_csv_to_cells, mesh_csv_to_attributes
from .dxf_reader import dxf_from_file_to_vertex, dxf_from_stream_to_vertex, DXFEntityType
from subsurface.core.reader_helpers.readers_data import GenericReaderFilesHelper, SupportedFormats
import numpy as np

__all__ = ['read_mesh_file_to_vertex', 'read_mesh_file_to_cells',
           'read_mesh_file_to_attr', 'cells_from_delaunay']


def read_mesh_file_to_vertex(reader_args: GenericReaderFilesHelper) -> np.ndarray:
    if reader_args.format is SupportedFormats.CSV:
        vertex = mesh_csv_to_vertex(reader_args.file_or_buffer, reader_args.columns_map,
                                    **reader_args.pandas_reader_kwargs)
    elif reader_args.format is SupportedFormats.DXF:
        vertex = dxf_from_file_to_vertex(
            file_path=reader_args.file_or_buffer,
            entity_type=reader_args.additional_reader_kwargs.get('entity_type', DXFEntityType.ALL)
        )
    elif reader_args.format is SupportedFormats.DXFStream:
        vertex = dxf_from_stream_to_vertex(reader_args.file_or_buffer)
    else:
        raise ValueError(f"Subsurface is not able to read the following extension: {reader_args.format}")
    return vertex


def read_mesh_file_to_cells(reader_args: GenericReaderFilesHelper) -> np.ndarray:
    extension = reader_args.format

    if extension == SupportedFormats.CSV:
        cells = mesh_csv_to_cells(
            path_to_file=reader_args.file_or_buffer,
            columns_map=reader_args.columns_map,
            **reader_args.pandas_reader_kwargs
        )
    else:
        raise ValueError(f"Subsurface is not able to read the following extension: {extension}")
    return cells


def read_mesh_file_to_attr(reader_args: GenericReaderFilesHelper):
    extension = reader_args.format
    if extension == SupportedFormats.CSV:
        attr = mesh_csv_to_attributes(reader_args.file_or_buffer,
                                      reader_args.columns_map,
                                      **reader_args.pandas_reader_kwargs)
    else:
        raise ValueError(f"Subsurface is not able to read the following extension: {extension}")
    return attr


def cells_from_delaunay(vertex):
    import pyvista as pv
    a = pv.PolyData(vertex)
    b = a.delaunay_2d().faces
    cells = b.reshape(-1, 4)[:, 1:]
    return cells



from typing import Union

from pathlib import Path

from ..structs import StructuredData, UnstructuredData


def get_extension(path):
    try:
        p = Path(path)
        return p.suffix
    except TypeError:
        return False


def replace_outliers(base_data: Union[StructuredData, UnstructuredData], dim=0, perc=0.99, replace_for=None):
    """@Edoardo Guerreiro https://stackoverflow.com/questions/60816533/
     is-there-a-built-in-function-in-xarray-to-remove-outliers-from-a-dataset"""

    data = base_data.data
    # calculate percentile
    threshold = data[dim].quantile(perc)

    # find outliers and replace them with max among remaining values
    mask = data[dim].where(abs(data[dim]) <= threshold)
    if replace_for == 'max':
        max_value = mask.max().values
        # .where replace outliers with nan
        mask = mask.fillna(max_value)
    elif replace_for == 'min':
        min_value = mask.min().values
        # .where replace outliers with nan
        mask = mask.fillna(min_value)

    print(mask)
    data[dim] = mask

    return data



from subsurface import StructuredData


class OctreeMesh:
    """
    TODO: implement as Dom discussed with data frames to track the levels.
    """

    def __init__(self, data: StructuredData):
        raise NotImplementedError



import numpy as np
import pandas as pd
from dataclasses import dataclass
from typing import Hashable, Literal

from ._combine_trajectories import create_combined_trajectory, MergeOptions
from .collars import Collars
from .survey import Survey
from ...structs import LineSet


@dataclass
class BoreholeSet:
    """
    This module provides a class, `BoreholeSet`, that represents a collection of boreholes. It contains methods for accessing coordinate data for each lithology in the boreholes.
    
    Notes: 
        - Collars is defined as 1 UnstructuredData
        - Combined trajectory is defined as 1 UnstructuredData

    Classes:
        - `BoreholeSet`: Represents a collection of boreholes.

    Methods:
        - `__init__`: Initializes a new `BoreholeSet` object with the specified input parameters.
        - `get_top_coords_for_each_lith`: Returns a dictionary of top coordinates for each lithology in the boreholes.
        - `get_bottom_coords_for_each_lith`: Returns a dictionary of bottom coordinates for each lithology in the boreholes.

    Attributes:
        - `collars`: A `Collars` object representing the collar information for the boreholes.
        - `survey`: A `Survey` object representing the survey information for the boreholes.
        - `combined_trajectory`: A `LineSet` object representing the combined trajectory of the boreholes.

    Usage:
        ```
        borehole_set = BoreholeSet(collars, survey, merge_option)
        top_coords = borehole_set.get_top_coords_for_each_lith()
        bottom_coords = borehole_set.get_bottom_coords_for_each_lith()
        ```

    Note: The example usage code provided above is for demonstration purposes only. Please replace `collars`, `survey`, and `merge_option` with the actual input parameters when using the `BoreholeSet` class.

    """
    __slots__ = ['collars', 'survey', 'combined_trajectory']
    collars: Collars
    survey: Survey
    combined_trajectory: LineSet

    def __init__(self, collars: Collars, survey: Survey, merge_option: MergeOptions, slice_=slice(None)):

        new_collars = self._remap_collars_with_survey(collars, survey)

        self.collars = new_collars
        self.survey = survey
        self.combined_trajectory: LineSet = create_combined_trajectory(collars, survey, merge_option, slice_)

    @staticmethod
    def _remap_collars_with_survey(collars, survey):
        import pandas as pd
        # Create a DataFrame from your first list
        df1 = pd.DataFrame({'name': collars.ids, 'x': collars.data.vertex[:, 0], 'y': collars.data.vertex[:, 1], 'z': collars.data.vertex[:, 2]})
        df1 = df1.set_index('name')
        # Reindex to match the second list of names
        df_reindexed = df1.reindex(survey.well_id_mapper.keys())
        new_collars = Collars.from_df(df_reindexed)
        return new_collars

    def to_binary(self, path: str) -> bool:
        # I need to implement the survey to and then name the files accordingly
        bytearray_le_collars: bytes = self.collars.data.to_binary()
        bytearray_le_trajectory: bytes = self.combined_trajectory.data.to_binary()

        new_file = open(f"{path}_collars.le", "wb")
        new_file.write(bytearray_le_collars)

        new_file = open(f"{path}_trajectory.le", "wb")
        new_file.write(bytearray_le_trajectory)
        return True

    def get_top_coords_for_each_lith(self) -> dict[Hashable, np.ndarray]:
        merged_df = self._merge_vertex_data_arrays_to_dataframe()
        component_lith_arrays = {}
        for lith, group in merged_df.groupby('lith_ids'):
            lith = int(lith)
            first_vertices = group.groupby('well_id').first().reset_index()
            array = first_vertices[['X', 'Y', 'Z']].values
            component_lith_arrays[lith] = array

        return component_lith_arrays

    def get_bottom_coords_for_each_lith(self, group_by: Literal['component lith', 'lith_ids'] = 'lith_ids') -> dict[Hashable, np.ndarray]:
        """
        Retrieves the bottom coordinates for each lithological component or lith ID from
        the merged vertex data arrays.

        This function groups the merged data by either 'component lith' or 'lith_ids',
        then extracts the coordinates of the bottommost vertices for each well. It
        returns a dictionary where keys are either lithological component identifiers
        or lith IDs, and values are arrays of 3D coordinates representing the bottom
        vertices.

        Args:
            group_by (Literal['component lith', 'lith_ids']): Specifies the grouping
                column to use for lithological components. Acceptable values are either
                'component lith' or 'lith_ids'. Defaults to 'lith_ids'.

        Returns:
            dict[Hashable, np.ndarray]: A dictionary mapping the lithological component
            or lith ID to an array of 3D coordinates ([X, Y, Z]) corresponding to the
            bottom vertices for each well.

        Raises:
            ValueError: If no groups are found from the specified `group_by` column.
        """
        merged_df = self._merge_vertex_data_arrays_to_dataframe()
        component_lith_arrays = {}
        group = merged_df.groupby(group_by)

        if group.ngroups == 0:
            raise ValueError("No components found")
        for lith, group in group:
            lith = int(lith)
            first_vertices = group.groupby('well_id').last().reset_index()
            array = first_vertices[['X', 'Y', 'Z']].values
            component_lith_arrays[lith] = array

        return component_lith_arrays

    def _merge_vertex_data_arrays_to_dataframe(self):
        ds = self.combined_trajectory.data.data
        # Convert vertex attributes to a DataFrame for easier manipulation
        vertex_attrs_df = ds['vertex_attrs'].to_dataframe().reset_index()
        vertex_attrs_df = vertex_attrs_df.pivot(index='points', columns='vertex_attr', values='vertex_attrs').reset_index()
        # Convert vertex coordinates to a DataFrame
        vertex_df = ds['vertex'].to_dataframe().reset_index()
        vertex_df = vertex_df.pivot(index='points', columns='XYZ', values='vertex').reset_index()
        # Merge the attributes with the vertex coordinates
        merged_df = pd.merge(vertex_df, vertex_attrs_df, on='points')
        # Create a dictionary to hold the numpy arrays for each component lith
        return merged_df



import json
import numpy as np
import pandas as pd


class LiquidEarthMesh:
    def __init__(self, vertex=None, cells=None, attributes=None, points_attributes=None, data_attrs=None):
        self.vertex = vertex  # Expected to be a numpy array of shape (N, 3)
        self.cells = cells  # Expected to be a numpy array of shape (M, K)
        self.attributes = attributes
        self.points_attributes = points_attributes
        self.data_attrs = data_attrs if data_attrs is not None else {}

    def to_binary(self, order='F') -> bytes:
        body_ = self._to_bytearray(order)
        header_ = self._set_binary_header()
        header_json = json.dumps(header_)
        header_json_bytes = header_json.encode('utf-8')
        header_json_length = len(header_json_bytes)
        header_json_length_bytes = header_json_length.to_bytes(4, byteorder='little')
        file = header_json_length_bytes + header_json_bytes + body_
        return file

    def _set_binary_header(self):
        header = {
                "vertex_shape"     : self.vertex.shape if self.vertex is not None else [0, 0],
                "cell_shape"       : self.cells.shape if self.cells is not None else [0, 0],
                "cell_attr_shape"  : self.attributes.shape if not self.attributes.empty else [0, 0],
                "vertex_attr_shape": self.points_attributes.shape if not self.points_attributes.empty else [0, 0],
                "cell_attr_names"  : self.attributes.columns.tolist() if not self.attributes.empty else [],
                "cell_attr_types"  : self.attributes.dtypes.astype(str).tolist() if not self.attributes.empty else [],
                "vertex_attr_names": self.points_attributes.columns.tolist() if not self.points_attributes.empty else [],
                "vertex_attr_types": self.points_attributes.dtypes.astype(str).tolist() if not self.points_attributes.empty else [],
                "xarray_attrs"     : self.data_attrs
        }
        return header

    def _to_bytearray(self, order):
        parts = []
        if self.vertex is not None:
            vertex_bytes = self.vertex.astype('float32').tobytes(order)
            parts.append(vertex_bytes)
        if self.cells is not None:
            cells_bytes = self.cells.astype('int32').tobytes(order)
            parts.append(cells_bytes)
        if not self.attributes.empty:
            cell_attr_bytes = self.attributes.values.astype('float32').tobytes(order)
            parts.append(cell_attr_bytes)
        if not self.points_attributes.empty:
            vertex_attr_bytes = self.points_attributes.values.astype('float32').tobytes(order)
            parts.append(vertex_attr_bytes)
        bytearray_le = b''.join(parts)
        return bytearray_le

    @classmethod
    def from_binary(cls, binary_data, order='F'):
        # Read header length
        header_length_bytes = binary_data[:4]
        header_length = int.from_bytes(header_length_bytes, byteorder='little')
        # Read header
        header_json_bytes = binary_data[4:4 + header_length]
        header_json = header_json_bytes.decode('utf-8')
        header = json.loads(header_json)
        # Read body
        body = binary_data[4 + header_length:]
        offset = 0

        # Parse vertices
        vertex_shape = header['vertex_shape']
        if vertex_shape[0] > 0 and vertex_shape[1] > 0:
            num_vertices = np.prod(vertex_shape)
            num_bytes = num_vertices * 4  # float32
            vertex = np.frombuffer(body[offset:offset + num_bytes], dtype=np.float32, count=num_vertices)
            offset += num_bytes
            vertex = vertex.reshape(vertex_shape, order=order)
        else:
            vertex = None

        # Parse cells
        cell_shape = header['cell_shape']
        if cell_shape[0] > 0 and cell_shape[1] > 0:
            num_cells = np.prod(cell_shape)
            num_bytes = num_cells * 4  # int32
            cells = np.frombuffer(body[offset:offset + num_bytes], dtype=np.int32, count=num_cells)
            offset += num_bytes
            cells = cells.reshape(cell_shape, order=order)
        else:
            cells = None

        # Parse cell attributes
        attributes = pd.DataFrame()
        cell_attr_shape = header['cell_attr_shape']
        if cell_attr_shape[0] > 0 and cell_attr_shape[1] > 0:
            num_attrs = np.prod(cell_attr_shape)
            num_bytes = num_attrs * 4  # float32
            cell_attr_values = np.frombuffer(body[offset:offset + num_bytes], dtype=np.float32, count=num_attrs)
            offset += num_bytes
            cell_attr_values = cell_attr_values.reshape(cell_attr_shape, order=order)
            attr_names = header['cell_attr_names']
            attributes = pd.DataFrame(cell_attr_values, columns=attr_names)
        else:
            attributes = None

        # Parse vertex attributes
        points_attributes = pd.DataFrame()
        vertex_attr_shape = header['vertex_attr_shape']
        if vertex_attr_shape[0] > 0 and vertex_attr_shape[1] > 0:
            num_attrs = np.prod(vertex_attr_shape)
            num_bytes = num_attrs * 4  # float32
            vertex_attr_values = np.frombuffer(body[offset:offset + num_bytes], dtype=np.float32, count=num_attrs)
            offset += num_bytes
            vertex_attr_values = vertex_attr_values.reshape(vertex_attr_shape, order=order)
            attr_names = header['vertex_attr_names']
            points_attributes = pd.DataFrame(vertex_attr_values, columns=attr_names)
        else:
            points_attributes = None

        data_attrs = header.get('xarray_attrs', {})

        return cls(vertex=vertex, cells=cells, attributes=attributes, points_attributes=points_attributes, data_attrs=data_attrs)




import io
from typing import Union

from ....core.structs import TriSurf
from ._trimesh_reader import load_with_trimesh, trimesh_to_unstruct, TriMeshTransformations


def load_gltf_with_trimesh(path_to_glb: Union[str | io.BytesIO], coordinate_system: TriMeshTransformations) -> TriSurf:
    """
    load_obj_with_trimesh(path_to_glb, plot=False)

    Loads a 3D object file in .glb format using trimesh, processes it, and converts it into 
    a subsurface TriSurf object for further analysis or usage. Optionally, it allows 
    plotting the loaded object.

    Parameters
    ----------
    path_to_glb : str
        Path to the .glb file containing the 3D object.
    plot : bool, optional
        A flag indicating whether the loaded 3D object should be plotted. Defaults to False.

    Returns
    -------
    subsurface.TriSurf
        A TriSurf object representing the processed 3D surface geometry.
    """
    trimesh = load_with_trimesh(path_to_glb, file_type="glb", coordinate_system=coordinate_system, plot=False)
    trisurf = trimesh_to_unstruct(trimesh)
    return trisurf



"""
TODO: This is legacy code waiting to be updated to the new ideas

"""

import xarray as xr
import numpy as np

from subsurface import optional_requirements


class Seismic:
    def __init__(self, data: np.ndarray, *args, **kwargs):
        """Seismic data object based on xarray.DataArray.
        
        Args:
            data (Array): np.ndarray of the seismic cube / section.
        """
        self.n_shp = len(data.shape)
        self._xarray = xr.DataArray(self._flip(data), *args, **kwargs)

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return getattr(self, attr)
        return getattr(self._xarray, attr)
    
    def __getitem__(self, item):
        if isinstance(item, str):
            return self._xarray._getitem_coord(item)

        # preserve coordinates 
        cp = list(self._xarray.coords.items())  # parent coordinates
        coords = [(cp[i]) for i, it in enumerate(item) if not type(it) == int]
        return Seismic(self._xarray[item].data, coords=coords)
    
    def __repr__(self):
        return self._xarray.__repr__()
    
    def __str__(self):
        return "Seismic"

    def _flip(self, data: np.ndarray) -> np.ndarray:
        """Flip SEGY data to fit with numpy array orientation.

        Args:
            data (np.ndarray): Seismic data.

        Returns:
            (np.ndarray) Flipped seismic data.
        """
        if self.n_shp == 1:
            return data
        if self.n_shp == 2:
            return data
        if self.n_shp == 3:
            return np.flip(data, axis=2)

    def add_coords(self):
        """Ability to easily add physical coordinates."""
        raise NotImplementedError

    def to_segy(self, filepath: str) -> None:
        """Write given Seismic to SEGY file using segyio.tools.from_array().
        
        Args:
            filepath (str): Filepath for SEGY file.
        """
        segyio = optional_requirements.require_segyio()
        segyio.tools.from_array(filepath, self._xarray.data)

    def plot_(self):
        return xr.plot.plot._PlotMethods(self)

    def plot(self):
        if self.n_shp == 1:
            return _plot_1d(self)
        elif self.n_shp == 2:
            pass
            # TODO: plot seismic section using imshow for correct orientation
        elif self.n_shp >= 3:
            _plot_3d(self)

    def create_pyvista_grid(self) -> 'pyvista.grid.UniformGrid':
        """Generate UniformGrid object for 3D plotting of the seismic.

        Args:
            seismic (Seismic): Seismic object.

        Returns:
            (pv.grid.UniformGrid)
        """
        pv = optional_requirements.require_pyvista()
        grid = pv.UniformGrid()
        grid.spacing = (1, 1, 1)  # TODO: cell sizes? vertical exaggeration etc
        grid.dimensions = np.array(self.data.shape) + 1
        grid.cell_data["values"] = self.data.flatten(order="F")
        # TODO: correct orientation of cube
        return grid

    def plot_3d_slices(self):
        pv = optional_requirements.require_pyvista()
        if self.n_shp != 3:
            raise AssertionError("Seismic data needs to be a 3-D volume.")
        # TODO: cmap, kwarg passthrough
        pv.OrthogonalSlicer(self.grid)


def _plot_1d(seismic: Seismic, linekwargs={}, fillkwargs={}):
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots(figsize=(2,8))
    
    lkwargs = dict(
        color="black",
        linewidth=1
    )
    lkwargs.update(linekwargs)

    fkwargs = dict(
        color="grey"
    )
    fkwargs.update(fillkwargs)

    y = np.arange(0, *seismic.data.shape)
    ax.plot(seismic.data, y, **lkwargs)
    x1 = seismic.data.copy()
    x1[x1<=0] = 0
    ax.fill_betweenx(y, x1=x1, **fkwargs)
    return ax


def _plot_3d(seismic: Seismic):
    if not hasattr(seismic, "grid"):
        seismic.grid = seismic.create_pyvista_grid()

    seismic.grid.plot(cmap="seismic")


def _plot_2d(seismic: Seismic):
    pass


def _plot_hist(seismic: Seismic):
    pass
        

def from_segy(filepath:str, coords=None) -> Seismic:
    """Create a Seismic data object from a SEGY file.
    
    Args:
        filepath (str): Filepath to the SEGY file.
    
    Returns:
        Seismic: Seismic data object based on xarray.DataArray.
    """
    segyio = optional_requirements.require_segyio()
    with segyio.open(filepath) as sf:
        sf.mmap()  # memory mapping
        xlines = sf.xlines
        ilines = sf.ilines
        samples = sf.samples
        header = sf.bin

    if not coords:
        coords = [
            ("ilines", ilines), 
            ("xlines", xlines),
            ("samples", samples)
        ]

    cube = segyio.tools.cube(filepath)
    seismic = Seismic(cube, coords=coords)
    seismic.header = header
    return seismic



import pandas as pd
import warnings
from typing import Dict

from subsurface.reader.readers_data import ReaderFilesHelper, ReaderWellsHelper, SupportedFormats
from subsurface.reader.wells.wells_utils import add_tops_from_base_and_altitude_in_place


def read_borehole_files(reader_wells_helper: ReaderWellsHelper) -> Dict[str, pd.DataFrame]:
    data_frames = dict()

    data_frames['collar_df'] = read_collar(reader_wells_helper.reader_collars_args)

    data_frames['survey_df'] = read_survey(reader_wells_helper.reader_survey_args)

    if reader_wells_helper.reader_lith_args is not None:
        data_frames['lith_df'] = read_lith(reader_wells_helper.reader_lith_args)

    if reader_wells_helper.reader_attr_args is not None:
        attributes_ = list()
        for e in reader_wells_helper.reader_attr_args:
            attributes_.append(read_attributes(e))
        data_frames['attrib_dfs'] = attributes_

    return data_frames


def read_collar(reader_helper: ReaderFilesHelper) -> pd.DataFrame:
    if reader_helper.usecols is None: reader_helper.usecols = [0, 1, 2, 3]
    if reader_helper.index_col is False: reader_helper.index_col = 0

    # Check file_or_buffer type
    d = check_format_and_read_to_df(reader_helper)
    map_rows_and_cols_inplace(d, reader_helper)

    return d


def read_survey(reader_helper: ReaderFilesHelper):
    if reader_helper.index_col is False: reader_helper.index_col = 0

    d = check_format_and_read_to_df(reader_helper)
    map_rows_and_cols_inplace(d, reader_helper)

    d_no_singles = _validate_survey_data(d)

    return d_no_singles


def read_lith(reader_helper: ReaderFilesHelper):
    """Columns MUST contain:
        - top
        - base
        - component lith
    """
    if reader_helper.index_col is False: reader_helper.index_col = 0

    d = check_format_and_read_to_df(reader_helper)
    map_rows_and_cols_inplace(d, reader_helper)
    lith_df = _validate_lith_data(d, reader_helper)

    return lith_df


def read_attributes(reader_helper: ReaderFilesHelper) -> pd.DataFrame:
    if reader_helper.index_col is False: reader_helper.index_col = 0

    d = check_format_and_read_to_df(reader_helper)

    if reader_helper.columns_map is not None: d.rename(reader_helper.columns_map, axis=1, inplace=True)
    if reader_helper.drop_cols is not None: d.drop(reader_helper.drop_cols, axis=1, inplace=True)

    _validate_attr_data(d)
    return d


def read_survey_df_from_las(reader_helper: ReaderFilesHelper, well_name: str) -> pd.DataFrame:
    """
    Reads a las file and returns a dataframe.
    
    """
    from subsurface.reader.wells._welly_reader import _create_welly_well_from_las
    welly_well = _create_welly_well_from_las(well_name, reader_helper.file_or_buffer)
    survey_df = welly_well.df()[reader_helper.usecols]
    map_rows_and_cols_inplace(survey_df, reader_helper)
    survey_df["well_name"] = well_name
    survey_df.set_index("well_name", inplace=True)
    return survey_df


def read_assay_df_from_las(reader_helper: ReaderFilesHelper, well_name: str) -> pd.DataFrame:
    from subsurface.reader.wells._welly_reader import _create_welly_well_from_las
    welly_well = _create_welly_well_from_las(well_name, reader_helper.file_or_buffer)
    assay_df = welly_well.df()
    assay_df["well_name"] = well_name
    assay_df.set_index("well_name", inplace=True)
    return assay_df


def check_format_and_read_to_df(reader_helper: ReaderFilesHelper) -> pd.DataFrame:
    if reader_helper.format == ".json":
        d = pd.read_json(reader_helper.file_or_buffer, orient='split')
    elif reader_helper.is_file_in_disk:
        reader = _get_reader(reader_helper.format)
        d = reader(reader_helper.file_or_buffer, **reader_helper.pandas_reader_kwargs)
    elif reader_helper.is_bytes_string:
        reader = _get_reader('.csv')
        d = reader(reader_helper.file_or_buffer, **reader_helper.pandas_reader_kwargs)
    elif reader_helper.is_python_dict:
        reader = _get_reader('dict')
        d = reader(reader_helper.file_or_buffer)
    else:
        raise AttributeError('file_or_buffer must be either a path or a dict')

    if type(d.columns) is str:  d.columns = d.columns.str.strip()  # Remove spaces at the beginning and end
    if type(d.index) is str: d.index = d.index.str.strip()  # Remove spaces at the beginning and end
    return d




def _get_reader(file_format):
    if file_format == SupportedFormats.XLXS:
        reader = pd.read_excel
    elif file_format == 'dict':
        reader = _dict_reader
    elif file_format == SupportedFormats.CSV:
        reader = pd.read_csv
    elif file_format == SupportedFormats.JSON:
        reader = _dict_reader
    else:
        raise ValueError(f"Subsurface is not able to read the following extension: {file_format}")
    return reader


def _dict_reader(dict_):
    """

    Args:
        dict_: data, index, columns

    """
    return pd.DataFrame(data=dict_['data'],
                        columns=dict_['columns'],
                        index=dict_['index'])




def _validate_lith_data(d: pd.DataFrame, reader_helper: ReaderFilesHelper) -> pd.DataFrame:
    given_top = pd.np.isin(['top', 'base', 'component lith'], d.columns).all()
    given_altitude_and_base = pd.np.isin(['altitude', 'base', 'component lith'], d.columns).all()

    if given_altitude_and_base and not given_top:
        d = add_tops_from_base_and_altitude_in_place(d, reader_helper.index_col, 'base', 'altitude')
    elif not given_top and not given_altitude_and_base:
        raise ValueError('basis column must be present in the file. Use '
                         'columns_map to assign column names to these fields.')
    lith_df = d[['top', 'base', 'component lith']]
    return lith_df


def _validate_attr_data(d):
    assert d.columns.isin(['basis']).any(), 'basis column' \
                                            'must be present in the file.' \
                                            'Use columns_map to assign' \
                                            'column names to these fields.'



from typing import List

from subsurface.writer.to_rex.common import file_header_size, encode
from subsurface.writer.to_rex.data_struct import RexLineSet, RexMesh, RexMaterial
from subsurface.writer.to_rex.material_encoder import material_encode
from subsurface.writer.to_rex.mesh_encoder import mesh_encode


__all__ = ['numpy_to_rex', 'w_data_blocks', 'w_block_data_type',
           'w_file_header_and_coord_system_block',
           'w_file_header_and_coord_system_block', 'write_rex_file',
           'read_rex_file']


def numpy_to_rex(
        rex_line_set: List[RexLineSet] = None,
        rex_meshes: List[RexMesh] = None,
        rex_material: List[RexMaterial] = None
):
    if rex_line_set is None:
        rex_line_set = list()
    if rex_meshes is None:
        rex_meshes = list()
    if rex_material is None:
        rex_material = list()

    data_block_bytes = bytearray()
    byte_size = 0

    data_block_bytes, data_id = w_data_blocks(rex_meshes, rex_material)

    n_data_blocks = data_id
    header_and_coord_block_bytes = w_file_header_and_coord_system_block(
        n_data_blocks=n_data_blocks,
        size_data_blocks=len(data_block_bytes),
        start_data=file_header_size
    )

    return header_and_coord_block_bytes + data_block_bytes


def w_data_blocks(rex_meshes: List[RexMesh], rex_material: List[RexMaterial]):
    data_id = 0

    rmesh_bytes, data_id = w_block_data_type(mesh_encode, rex_meshes, data_id)
    rmaterial_bytes, data_id = w_block_data_type(material_encode, rex_material,
                                                 data_id)

    blocks_bytes = rmesh_bytes + rmaterial_bytes
    return blocks_bytes, data_id


def w_block_data_type(encoder, rex_objects: List, data_id: int):
    data_block_bytes = bytearray()
    for rex_obj in rex_objects:
        data_block_bytes += encoder(rex_obj, data_id=data_id)
        data_id += 1

    return data_block_bytes, data_id


def w_file_header_and_coord_system_block():
    return


def w_file_header_and_coord_system_block(n_data_blocks, size_data_blocks, version=1,
                                         start_data=86, srid=3876, offsets=None):
    """
    Function that writes the header block of a rexfile:

    Args:
        n_data_blocks:
        size_data_blocks:
        version (int): Version of the file
        start_data (int): Position where data start. This is after the header
         and coordinate system. If everything works fine it should be 86
        srid (int): Spatial reference system identifier (srid)
        offsets:

    Returns:

    """
    reserved = '0' * 42
    if offsets is None:
        offsets = [0, 0, 0]

    input_ = [('REX1', 'bytes'),  # REX1
              (version, 'uint16'),  # file version
              (0, 'uint32'),  # CRC32
              (n_data_blocks, 'uint16'),  # Number of DATA BLOCKS
              (start_data, 'uint16'),  # StartData
              (size_data_blocks, 'uint64'),  # Size of all data blocks
              (reserved, 'bytes'),  # Reserved
              # Coordinate system block
              (srid, 'uint32'),  # Spatial reference system identifier (srid)
              (4, 'uint16'),  # Size of the name of the used system.
              ('EPSG', 'bytes'),  # name of the used system.
              (offsets, 'float32')]  # Global x, y, z offset

    block_bytes = encode(input_)
    return block_bytes


def write_rex_file(bytes, path: str):
    """Write to disk a rexfile from its binary format"""

    new_file = open(path + ".rex", "wb")
    new_file.write(bytes)
    return True


def read_rex_file(path: str) -> bytes:
    with open(path, "rb") as f:
        bytes_read = f.read()
    return bytes_read



import json


def base_structs_to_binary_file(path, base_struct, order='F'):
    try:
        bytearray_le, header = base_struct.default_data_array_to_binary_legacy(order=order)
    except AttributeError:
        bytearray_le, header = base_struct.to_binary_legacy(order=order)
    with open(path+'.json', 'w') as outfile:
        json.dump(header, outfile)

    new_file = open(path+".le", "wb")
    new_file.write(bytearray_le)


from typing import Hashable, Optional

import numpy as np
import pandas as pd

from subsurface import optional_requirements
from ...structs.base_structures import UnstructuredData


def data_frame_to_unstructured_data(survey_df: 'pd.DataFrame', number_nodes: int, attr_df: Optional['pd.DataFrame'] = None,
                                     duplicate_attr_depths: bool = False) -> UnstructuredData:
    wp = optional_requirements.require_wellpathpy()

    cum_vertex: np.ndarray = np.empty((0, 3), dtype=np.float32)
    cells: np.ndarray = np.empty((0, 2), dtype=np.int_)
    cell_attr: pd.DataFrame = pd.DataFrame(columns=['well_id'], dtype=np.float32)
    vertex_attr: pd.DataFrame = pd.DataFrame()

    for e, (borehole_id, data) in enumerate(survey_df.groupby(level=0)):
        dev = wp.deviation(
            md=data['md'].values,
            inc=data['inc'].values,
            azi=data['azi'].values
        )

        md_min = dev.md.min()
        md_max = dev.md.max()

        attr_depths = _grab_depths_from_attr(
            attr_df=attr_df,
            borehole_id=borehole_id,
            duplicate_attr_depths=duplicate_attr_depths,
            md_max=md_max,
            md_min=md_min
        )

        # Now combine attr_depths with depths
        md_min = dev.md.min()
        md_max = dev.md.max()
        depths = np.linspace(md_min, md_max, number_nodes)
        depths = np.union1d(depths, attr_depths)
        depths.sort()

        # Resample positions at depths
        pos = dev.minimum_curvature().resample(depths=depths)
        vertex_count = cum_vertex.shape[0]

        this_well_vertex = np.vstack([pos.easting, pos.northing, pos.depth]).T
        cum_vertex = np.vstack([cum_vertex, this_well_vertex])
        measured_depths = _calculate_distances(array_of_vertices=this_well_vertex)

        n_vertex_shift_0 = np.arange(0, len(pos.depth) - 1, dtype=np.int_)
        n_vertex_shift_1 = np.arange(1, len(pos.depth), dtype=np.int_)
        cell_per_well = np.vstack([n_vertex_shift_0, n_vertex_shift_1]).T + vertex_count
        cells = np.vstack([cells, cell_per_well])

        attribute_values = np.isin(depths, attr_depths)

        vertex_attr_per_well = pd.DataFrame({
                'well_id'        : [e] * len(pos.depth),
                'measured_depths': measured_depths,
                'is_attr_point'  : attribute_values,
        })

        vertex_attr = pd.concat([vertex_attr, vertex_attr_per_well], ignore_index=True)

        # Add the id (e), to cell_attr
        cell_attr = pd.concat([cell_attr, pd.DataFrame({'well_id': [e] * len(cell_per_well)})], ignore_index=True)

    unstruct = UnstructuredData.from_array(
        vertex=cum_vertex,
        cells=cells.astype(int),
        vertex_attr=vertex_attr.reset_index(drop=True),
        cells_attr=cell_attr.reset_index(drop=True)
    )

    unstruct.data.attrs["well_id_mapper"] = {well_id: e for e, well_id in enumerate(survey_df.index.unique(level=0))}

    return unstruct


def _grab_depths_from_attr(
        attr_df: pd.DataFrame,
        borehole_id: Hashable,
        duplicate_attr_depths: bool,
        md_max: float,
        md_min: float
) -> np.ndarray:
    # Initialize attr_depths and attr_labels as empty arrays
    attr_depths = np.array([], dtype=float)
    attr_labels = np.array([], dtype='<U4')  # Initialize labels for 'top' and 'base'

    if attr_df is None or ("top" not in attr_df.columns and "base" not in attr_df.columns):
        return attr_depths

    try:
        vals = attr_df.loc[borehole_id]

        tops = np.array([], dtype=float)
        bases = np.array([], dtype=float)

        if 'top' in vals:
            if isinstance(vals, pd.DataFrame):
                tops = vals['top'].values.flatten()
            else:
                tops = np.array([vals['top']])
            # Convert to float and remove NaNs
            tops = tops.astype(float)
            tops = tops[~np.isnan(tops)]
            # Clip to within md range
            tops = tops[(tops >= md_min) & (tops <= md_max)]

        if 'base' in vals:
            if isinstance(vals, pd.DataFrame):
                bases = vals['base'].values.flatten()
            else:
                bases = np.array([vals['base']])
            # Convert to float and remove NaNs
            bases = bases.astype(float)
            bases = bases[~np.isnan(bases)]
            # Clip to within md range
            bases = bases[(bases >= md_min) & (bases <= md_max)]

        # Combine tops and bases into attr_depths with labels
        attr_depths = np.concatenate((tops, bases))
        attr_labels = np.array(['top'] * len(tops) + ['base'] * len(bases))

        # Drop duplicates while preserving order
        _, unique_indices = np.unique(attr_depths, return_index=True)
        attr_depths = attr_depths[unique_indices]
        attr_labels = attr_labels[unique_indices]

    except KeyError:
        # No attributes for this borehole_id or missing columns
        attr_depths = np.array([], dtype=float)
        attr_labels = np.array([], dtype='<U4')

    # If duplicate_attr_depths is True, duplicate attr_depths with a tiny offset
    if duplicate_attr_depths and len(attr_depths) > 0:
        tiny_offset = (md_max - md_min) * 1e-6  # A tiny fraction of the depth range
        # Create offsets: +tiny_offset for 'top', -tiny_offset for 'base'
        offsets = np.where(attr_labels == 'top', tiny_offset, -tiny_offset)
        duplicated_attr_depths = attr_depths + offsets
        # Ensure the duplicated depths are within the md range
        valid_indices = (duplicated_attr_depths >= md_min) & (duplicated_attr_depths <= md_max)
        duplicated_attr_depths = duplicated_attr_depths[valid_indices]
        # Original attribute depths
        original_attr_depths = attr_depths
        # Combine originals and duplicates
        attr_depths = np.hstack([original_attr_depths, duplicated_attr_depths])

    return attr_depths


def _calculate_distances(array_of_vertices: np.ndarray) -> np.ndarray:
    # Calculate the differences between consecutive points
    differences = np.diff(array_of_vertices, axis=0)

    # Calculate the Euclidean distance for each pair of consecutive points
    distances = np.linalg.norm(differences, axis=1)
    # Insert a 0 at the beginning to represent the starting point at the surface
    measured_depths = np.insert(np.cumsum(distances), 0, 0)
    return measured_depths



from subsurface import StructuredData


class StructuredSurface:
    def __init__(self, structured_data: StructuredData):
        # TODO check structured_data has two coordinates
        self.ds = structured_data

    # Add pyvista methods of gridded data



from typing import Iterable, Union, List

import numpy as np
import pandas as pd

try:
    import welly
    from welly import Well, Location, Project, Curve
    from striplog import Striplog, Component

    welly_imported = True
except ImportError:
    welly_imported = False

__all__ = ['WellyToSubsurfaceHelper', ]


class WellyToSubsurfaceHelper:
    def __init__(
            self,
            collar_df: pd.DataFrame = None,
            survey_df: pd.DataFrame = None,
            lith_df: pd.DataFrame = None,
            attrib_dfs: List[pd.DataFrame] = None
    ):
        """ Class that wraps `welly` to read borehole data - las files (upcoming)
         and deviations, csv, excel - and converts it into a
         `subsurface.UnstructuredData`

        This class is only meant to be extended with all the necessary functionality
         to load borehole data. For extensive manipulations of the data
         it should be done in `welly` itself.

        A borehole has:

            -  Datum (XYZ location)

            -  Deviation

            - Lithology: For this we are going to need striplog

            - Logs

        Everything would be a LineSet with a bunch of properties

        """

        if welly_imported is False:
            raise ImportError('You need to install welly to read well data.')

        self.welly_project = Project([])
        self._well_names = set()
        self._unique_formations = None

        if collar_df is not None: self.add_datum(collar_df)
        if survey_df is not None: self.add_deviation(survey_df)
        if lith_df is not None:  self.add_striplog(lith_df)

        # First check if is just a path or list
        if attrib_dfs is not None:
            for e, attrib in enumerate(attrib_dfs):
                self.add_assays(attrib, basis='basis')

    def __repr__(self):
        return self.p.__repr__()

    @property
    def p(self):
        """Project Alias"""
        return self.welly_project

    @p.setter
    def p(self, p):
        self.welly_project = p

    @property
    def lith_component_table(self):
        return [Component({'lith': l}) for l in self._unique_formations]

    @lith_component_table.setter
    def lith_component_table(self, unique_formations):
        self._unique_formations = unique_formations

    def add_wells(self, well_names: Iterable):
        new_boreholes = set(well_names).difference(self._well_names)

        self._well_names = self._well_names.union(well_names)
        for b in new_boreholes:
            # TODO: Name and uwi should be different
            if welly.__version__ < '0.5':
                w = Well(params={'header': {'name': b, 'uwi': b}})
            else:
                w = Well()
                w.uwi = b
                w.name = b
            # w.location = Location(params={'kb': 100})
            self.p += w
        return self.p

    def add_datum(self, data: pd.DataFrame):
        self.add_wells(
            well_names=np.unique(data.index)
        )

        for b in np.unique(data.index):
            w = self.welly_project.get_well(b)
            datum = data.loc[[b]]
            assert datum.shape[1] == 3, 'datum must be XYZ coord'

            w.location.position = datum.values[0]

        return self.welly_project

    def add_collar(self, data: pd.DataFrame):
        """Alias for add_datum"""
        return self.add_datum(data=data)

    def add_striplog(self, data: pd.DataFrame):
        unique_borehole = np.unique(data.index)
        self.add_wells(unique_borehole)
        self.lith_component_table = data['component lith'].unique()
        missed_borehole = []
        for b in unique_borehole:
            w = self.p.get_well(b)
            data_dict = data.loc[[b]].to_dict('list')
            data_csv = data.loc[[b]].to_csv()
            # s = Striplog.from_dict_advanced(data_dict, points=True)
            # s = Striplog.from_dict(data_dict)
            s = Striplog.from_csv(text=data_csv)

            try:
                n_basis = w.location.md.shape[0]
            except TypeError:
                n_basis = 2
            try:
                start, stop, step_size = self._calculate_basis_parameters(
                    w,
                    n_basis)
                s_log, basis, table = s.to_log(step_size, start, stop,
                                               return_meta=True)

                w.data['lith'] = s
                w.data['lith_log'] = Curve(s_log, basis)
            except TypeError:
                missed_borehole.append(b)
                continue

        print('The following striplog failed being processed: ', missed_borehole)

        return self.p

    def add_assays(self, data: pd.DataFrame, basis: Union[str, Iterable]):
        unique_borehole = np.unique(data.index)
        self.add_wells(unique_borehole)
        assay_attributes = data.columns

        if type(basis) == str:
            assay_attributes = assay_attributes.drop(basis)
            basis = data[basis]
        elif type(basis) == Iterable:
            pass
        else:
            raise AttributeError('basis must be either a string with the column name'
                                 'or a array like object')

        for b in unique_borehole:
            for a in assay_attributes:
                w = self.p.get_well(b)
                w.data[a] = Curve(
                    data=data[a],
                    basis=basis
                )

        return self.p

    @staticmethod
    def _calculate_basis_parameters(well, n_points):

        max_ = well.location.md.max()
        min_ = well.location.md.min()
        step_size = (max_ - min_) / n_points
        return min_ + step_size / 2, max_ - step_size / 2, step_size + 1e-12

    def add_deviation(
            self,
            deviations: pd.DataFrame,
            td=None,
            method='mc',
            update_deviation=True,
            azimuth_datum=0
    ):
        """ Add a deviation survey to this instance, and try to compute a position
         log from it.

        """
        unique_borehole = np.unique(deviations.index)
        self.add_wells(unique_borehole)

        for b in unique_borehole:
            w = self.p.get_well(b)
            deviations_df: pd.DataFrame = deviations.loc[[b], ['md', 'inc', 'azi']]
            deviations_df.fillna(0, inplace=True)
            w.location.add_deviation(
                deviation=deviations_df,
                td=td,
                method=method,
                update_deviation=update_deviation,
                azimuth_datum=azimuth_datum)
            if w.location.position is None:
                raise ValueError('Deviations could not be calculated.')

        return self.p



import io

import numpy as np

import subsurface
from ....optional_requirements import require_omf, require_pyvista
from ....core.structs.base_structures import UnstructuredData


def omf_stream_to_unstructs(stream: io.BytesIO) -> list[UnstructuredData]:
    pyvista = require_pyvista()
    omfvista = require_omf()
    omf = omfvista.load_project(stream)

    all_vertex = []
    all_cells = []
    cell_attr = []
    _last_cell: int = 0

    for i in range(omf.n_blocks):
        block: pyvista.PolyData = omf[i]
        cell_type = block.get_cell(0).type

        if cell_type == pyvista.CellType.TRIANGLE:
            pyvista_unstructured_grid: pyvista.UnstructuredGrid = block.cast_to_unstructured_grid()
            all_vertex.append(pyvista_unstructured_grid.points)
            cells: np.ndarray = pyvista_unstructured_grid.cells.reshape(-1, 4)[:, 1:]
            if len(all_cells) > 0:
                cells = cells + _last_cell

            all_cells.append(cells)
            cell_attr.append(np.ones(len(all_cells[-1])) * i)
            _last_cell = cells.max() + 1

    # * Create the unstructured data
    import pandas as pd

    unstructured_data = subsurface.UnstructuredData.from_array(
        vertex=np.vstack(all_vertex),
        cells=np.vstack(all_cells),
        cells_attr=pd.DataFrame(np.hstack(cell_attr), columns=["Block id"]),
    )
    return [unstructured_data]



import os
from io import BytesIO
from typing import Union

from subsurface.core.structs import StructuredData

from .... import optional_requirements
from ....core.structs import UnstructuredData
from subsurface.core.reader_helpers.readers_data import GenericReaderFilesHelper
import numpy as np
import pandas as pd


def read_VTK_structured_grid(file_or_buffer: Union[str, BytesIO], active_scalars: str) -> StructuredData:
    pv = optional_requirements.require_pyvista()

    if isinstance(file_or_buffer, BytesIO):
        # If file_or_buffer is a BytesIO, write it to a temporary file
        from tempfile import NamedTemporaryFile
        with NamedTemporaryFile('wb', suffix='.vtk', delete=False) as temp_file:
            # Write the BytesIO content to the temporary file
            getvalue: bytes = file_or_buffer.getvalue()
            temp_file.write(getvalue)
            temp_file.flush()  # Make sure all data is written
            temp_file_name = temp_file.name  # Store the temporary file name
        try:
            # Use pyvista.read() to read from the temporary file
            pyvista_obj = pv.read(temp_file_name)
        finally:
            # Ensure the temporary file is deleted after reading
            os.remove(temp_file_name)
    else:
        # If it's a file path, read directly
        pyvista_obj = pv.read(file_or_buffer)

    try:
        struct: StructuredData = StructuredData.from_pyvista(
            pyvista_object=pyvista_obj,
            data_array_name=active_scalars
        )
    except Exception as e:
        raise ValueError(f"Failed to convert to StructuredData: {e}")

    return struct


def read_volumetric_mesh_to_subsurface(reader_helper_coord: GenericReaderFilesHelper,
                                       reader_helper_attr: GenericReaderFilesHelper) -> UnstructuredData:
    df_coord = read_volumetric_mesh_coord_file(reader_helper_coord)
    if len(df_coord.columns) == 1:
        raise ValueError(
            "The attributes file has only one column, probably the columns are not being separated correctly. Use 'sep' in Additional Reader Arguments"
        )

    df_attr = read_volumetric_mesh_attr_file(reader_helper_attr)
    # Check if there are more than one column and if it is only one raise an error that probably the columns have not been properly separated. Use "sep" in Additional Reader Arguments
    if len(df_attr.columns) == 1:
        raise ValueError(
            "The attributes file has only one column, probably the columns are not being separated correctly. Use 'sep' in Additional Reader Arguments"
        )
    
    combined_df = df_coord.merge(df_attr, left_index=True, right_index=True)
    ud = UnstructuredData.from_array(
        vertex=combined_df[['x', 'y', 'z']], cells="points",
        attributes=combined_df[['pres', 'temp', 'sg', 'xco2']]
    )
    return ud


def read_volumetric_mesh_coord_file(reader_helper: GenericReaderFilesHelper) -> pd.DataFrame:
    df = pd.read_csv(
        filepath_or_buffer=reader_helper.file_or_buffer,
        **reader_helper.pandas_reader_kwargs
    )
    if reader_helper.columns_map is not None:
        df.rename(
            mapper=reader_helper.columns_map,
            axis="columns",
            inplace=True
        )

    df.dropna(axis=0, inplace=True)

    df.x = df.x.astype(float)
    df.y = df.y.astype(float)
    df.z = df.z.astype(float)
    # Throw error if empty
    if df.empty:
        raise ValueError("The file is empty")
    
    return df


def read_volumetric_mesh_attr_file(reader_helper: GenericReaderFilesHelper) -> pd.DataFrame:
    df = pd.read_table(reader_helper.file_or_buffer, **reader_helper.pandas_reader_kwargs)
    df.columns = df.columns.astype(str).str.strip()
    return df


def pv_cast_to_structured_grid(pyvista_object: 'pv.DataSet') -> 'pv.StructuredGrid':
    pv = optional_requirements.require_pyvista()

    match pyvista_object:
        case pv.UnstructuredGrid():
            # In a previous version of Subsurface there was an ill-formed attempt at supporting some unstructured grids here.
            # I've left this function to minimize downstream changes and also in case we decide to revive anything in that direction.
            raise ValueError("Cannot generally convert unstructured grids to structured grids.")
        case _:
            return pyvista_object.cast_to_structured_grid()



import enum
from dataclasses import dataclass
from typing import Dict, List, Tuple, Union, Literal

import numpy as np
import xarray as xr

from ....optional_requirements import require_pyvista


class StructuredDataType(enum.Enum):
    REGULAR_AXIS_ALIGNED = 0  #: Regular axis aligned grid. Distance between consecutive points is constant
    REGULAR_AXIS_UNALIGNED = 1  #: Regular axis unaligned grid. Distance between consecutive points is constant
    IRREGULAR_AXIS_ALIGNED = 2  #: Irregular axis aligned grid. Distance between consecutive points is not constant
    IRREGULAR_AXIS_UNALIGNED = 3  #: Irregular axis unaligned grid. Distance between consecutive points is not constant

    # [CLN] This terminology looks odd to me.
    # "Uniform" vs. "non-uniform" is what PyVista uses instead of "regular" vs. "irregular".
    # "Rectilinear" vs. "curvilinear" is what Pyvista uses instead of "axis-aligned" vs "axis-unaligned".
    # As of right now it's not clear to me that anything other than REGULAR_AXIS_ALIGNED is actually valid in this file.

@dataclass(frozen=False)
class StructuredData:
    data: xr.Dataset
    _active_data_array_name: str = "data_array"
    type: StructuredDataType = StructuredDataType.REGULAR_AXIS_ALIGNED
    dtype: Literal["float32", "float64"] = "float32"

    """Primary structure definition for structured data

       Check out other constructors: `StructuredData.from_numpy`,
        `StructuredData.from_data_array` and `StructuredData.from_dict`

    Args:
        data (xr.Dataset): object containing
         structured data, i.e. data that can be stored in multidimensional
         numpy array. The preferred type to pass as data is directly a
         xr.Dataset to be sure all the attributes are set and named as the user
         wants.
        data_array_name (str): If data is a numpy array or xarray DataArray, data_name
         provides the name for the xarray data variable
     
    Attributes:
        data (xarray.Dataset)
    """

    @property
    def active_data_array_name(self):
        data_var_list = list(self.data.data_vars.keys())
        if self._active_data_array_name not in data_var_list:
            raise ValueError("data_array_name not found in data_vars: {}".format(data_var_list))
        return self._active_data_array_name

    @active_data_array_name.setter
    def active_data_array_name(self, data_array_name: str):
        self._active_data_array_name = data_array_name

    @classmethod
    def from_numpy(cls, array: np.ndarray, coords: dict = None, data_array_name: str = "data_array",
                   dim_names: List[str] = None):
        if dim_names is None:
            dim_names = cls._default_dim_names(array.ndim)
        # if they are more than 3 we do not know the dimension name but it should valid:

        dataset: xr.Dataset = xr.Dataset(
            data_vars=
            {
                    data_array_name: (dim_names, array)
            },
            coords=coords
        )

        return cls(dataset, data_array_name)

    @classmethod
    def from_data_array(cls, data_array: xr.DataArray, data_array_name: str = "data_array"):
        dataset: xr.Dataset = xr.Dataset(
            data_vars={
                    data_array_name: data_array
            },
            coords=data_array.coords
        )

        return cls(dataset, data_array_name)

    @classmethod
    def from_dict(cls, data_dict: Dict[str, xr.DataArray], coords: Dict[str, str] = None, data_array_name: str = "data_array"):
        dataset: xr.Dataset = xr.Dataset(data_vars=data_dict, coords=coords)
        return cls(dataset, data_array_name)

    @classmethod
    def from_pyvista(
            cls,
            pyvista_object: 'pyvista.DataSet',
            data_array_name: str = "data_array"
    ):
        pyvista = require_pyvista()

        def rectilinear_is_uniform(
                rectilinear_grid: pyvista.RectilinearGrid,
                relative_tolerance: float = 1e-6,
                absolute_tolerance: float = 1e-12,
            ) -> bool:

            def axis_is_uniform(v: np.ndarray) -> bool:
                v = np.asarray(v, dtype=float)
                if v.size <= 2:
                    # 0, 1 or 2 points → treat as uniform for our purposes
                    return True
                diffs = np.diff(v)
                first = diffs[0]
                return np.allclose(diffs, first, rtol=relative_tolerance, atol=absolute_tolerance)

            return (axis_is_uniform(rectilinear_grid.x)
                and axis_is_uniform(rectilinear_grid.y)
                and axis_is_uniform(rectilinear_grid.z))

        extended_help_message = "Only uniform rectilinear grids are currently supported. The VTK format is developed by KitWare and you can use their free software ParaView to further inspect your file. In ParaView, in Information > Data Statistics, the Type must be Image (Uniform Rectilinear Grid). Furthermore, you can use ParaView to interpolate your data on to a uniform rectilinear grid and to export it as Image type."

        match pyvista_object:
            case pyvista.UnstructuredGrid():
                # In a previous version of Subsurface there was an ill-formed attempt at supporting some unstructured grids here.
                # I've left this function to minimize downstream changes and also in case we decide to revive anything in that direction.
                raise ValueError(f"Cannot generally convert unstructured grids to structured grids. {extended_help_message}")
            case pyvista.ImageData():
                pass
            case pyvista.RectilinearGrid() as rectilinear:
                if not rectilinear_is_uniform(rectilinear):
                    raise NotImplementedError(f"Non-uniform rectilinear grid conversion is not yet implemented. {extended_help_message}")
            case _:
                raise ValueError(f"Unexpected VTK grid type. {extended_help_message}")

        grid = pyvista_object.cast_to_structured_grid()

        # Extract p

        # Extract cell data and point data (if any)
        data_vars = {}

        # TODO: I need to do something with the bounds

        dimensions = np.array(grid.dimensions) - 1
        default_dim_names = cls._default_dim_names(dimensions.shape[0])

        bounds: tuple = grid.bounds
        coords = {}
        for i, dim in enumerate(default_dim_names):
            coords[dim] = np.linspace(
                start=bounds[i * 2],
                stop=bounds[i * 2 + 1],
                num=dimensions[i],
                endpoint=False
            )

        for name in grid.cell_data:
            cell_attr_data: pyvista.pyvista_ndarray = grid[name]
            cell_attr_data_reshaped = cell_attr_data.reshape(dimensions, order='F')

            data_vars[name] = xr.DataArray(
                data=cell_attr_data_reshaped,
                dims=default_dim_names,
                name=name
            )

        dataset: xr.Dataset = xr.Dataset(
            data_vars=data_vars,
            coords=coords
        )
        return cls(dataset, data_array_name)

    @property
    def values(self) -> np.ndarray:
        return self.data[self.active_data_array_name].values

    _bounds: Tuple[float, float, float, float, float, float] = None

    @property
    def bounds(self):
        if self._bounds is not None:
            return self._bounds

        array_: xr.DataArray = self.data[self.active_data_array_name]
        bounds = self._get_bounds_from_coord(array_)
        return bounds

    @bounds.setter
    def bounds(self, bounds: Tuple[float, float, float, float, float, float]):
        """
        Set the bounds of the structured data. This is useful for defining the
        spatial extent of the data in a structured grid.

        Args:
            bounds (Tuple[float, float, float, float, float, float]): A tuple containing
                the minimum and maximum values for each dimension (xmin, xmax, ymin, ymax, zmin, zmax).
        """
        self._bounds = bounds

    @property
    def shape(self):
        return self.active_data_array.shape

    @property
    def active_data_array(self):
        return self.data[self.active_data_array_name]

    @staticmethod
    def _get_bounds_from_coord(xr_obj: xr.DataArray):
        bounds = {}
        for coord in xr_obj.coords:
            bounds[coord] = (xr_obj[coord].min().item(), xr_obj[coord].max().item())
        return bounds

    def default_data_array_to_binary_legacy(self, order: Literal["K", "A", "C", "F"] = 'F'):
        bytearray_le = self._to_bytearray(order=order)
        header = self._set_binary_header()

        return bytearray_le, header

    def to_binary(self, order: Literal["K", "A", "C", "F"] = 'F') -> bytes:
        """Converts the structured data to a binary file
        
        Notes: 
            Only the active data array is converted to binary for now 
        """

        body_ = self._to_bytearray(order)
        header = self._set_binary_header()

        import json
        header_json = json.dumps(header)
        header_json_bytes = header_json.encode('utf-8')
        header_json_length = len(header_json_bytes)
        header_json_length_bytes = header_json_length.to_bytes(4, byteorder='little')
        file = header_json_length_bytes + header_json_bytes + body_
        return file

    def _set_binary_header(self) -> Dict:
        data_array = self.active_data_array

        match self.type:
            case StructuredDataType.REGULAR_AXIS_ALIGNED:
                header = {
                        "data_shape": self.shape,
                        "bounds"    : self.bounds,
                        "transform" : None,
                        "dtype"     : self.dtype,
                        "data_name" : self.active_data_array_name
                }
            case _:
                raise NotImplementedError(f"StructuredDataType {self.type} not implemented yet")

        return header

    def _to_bytearray(self, order: Literal["K", "A", "C", "F"]) -> bytes:
        data_array = self.active_data_array

        data = data_array.values.astype(self.dtype).tobytes(order)
        bytearray_le = data
        return bytearray_le

    @classmethod
    def _default_dim_names(cls, n_dims: int):
        if n_dims == 2:
            dim_names = ['x', 'y']
        elif n_dims == 3:
            dim_names = ['x', 'y', 'z']
        else:
            dim_names = ['dim' + str(i) for i in range(n_dims)]
        return dim_names

    def to_netcdf(self, path: str, **to_netcdf_kwargs):
        """
        Serializes the current StructuredData instance to a NetCDF file.

        Args:
            path (str): The path (including file name) where the NetCDF file will be saved.
            **to_netcdf_kwargs: Additional keyword arguments forwarded to xarray's `to_netcdf`.
        """
        # Copy the dataset (shallow copy of the data structure, no copying of the underlying arrays)
        ds = self.data.copy(deep=False)

        # Store relevant metadata as global attributes:
        ds.attrs["active_data_array_name"] = self._active_data_array_name
        ds.attrs["structured_data_type"] = self.type.name  # e.g., "REGULAR_AXIS_ALIGNED"
        ds.attrs["dtype"] = self.dtype  # e.g., "float32"

        # Use xarray's to_netcdf
        ds.to_netcdf(path, **to_netcdf_kwargs)

    @classmethod
    def from_netcdf(cls, path: str, **from_netcdf_kwargs):
        """
        Deserializes a NetCDF file into a StructuredData instance.

        Args:
            path (str): The path to the NetCDF file to read.
            **from_netcdf_kwargs: Additional keyword arguments forwarded to xarray's `open_dataset`.

        Returns:
            StructuredData: A new instance of StructuredData loaded from the file.
        """
        ds = xr.open_dataset(path, **from_netcdf_kwargs)

        # Retrieve what was stored in attrs (with defaults if missing)
        data_array_name = ds.attrs.get("active_data_array_name", "data_array")
        dtype_str: str = ds.attrs.get("dtype", "float32")
        if dtype_str not in ["float32", "float64"]:
            raise ValueError(f"Unsupported dtype: {dtype_str}")

        sdt_str = ds.attrs.get("structured_data_type", "REGULAR_AXIS_ALIGNED")

        # Convert strings back to your enum or any other type
        # (assuming StructuredDataType is an Enum where name matches sdt_str)
        if sdt_str not in StructuredDataType.__members__:
            raise ValueError(f"Unsupported structured_data_type: {sdt_str}")
        structured_data_type: StructuredDataType = StructuredDataType[sdt_str]

        return cls(
            data=ds,
            _active_data_array_name=data_array_name,
            type=structured_data_type,
            dtype=dtype_str
        )



import pandas as pd
from dataclasses import dataclass

from ...structs.base_structures import UnstructuredData
from ...structs.base_structures.base_structures_enum import SpecialCellCase
from ...structs.unstructured_elements import PointSet


@dataclass
class Collars:
    ids: list[str]
    collar_loc: PointSet
    
    @classmethod
    def from_df(cls, df: pd.DataFrame):
        unstruc: UnstructuredData = UnstructuredData.from_array(
            vertex=df[["x", "y", "z"]].values,
            cells=SpecialCellCase.POINTS
        )
        return cls(
            ids=df.index.to_list(),
            collar_loc=PointSet(data=unstruc)
        )
    
    @property
    def data(self):
        return self.collar_loc.data


import enum
import pathlib
import pandas as pd

from subsurface.core.utils.utils_core import get_extension
from pydantic import BaseModel, Field, model_validator
from typing import Union, List, Optional

if pd.__version__ < '1.4.0':
    pass
elif pd.__version__ >= '1.4.0':
    from pandas._typing import FilePath, ReadCsvBuffer

    fb = Union[FilePath, ReadCsvBuffer[bytes], ReadCsvBuffer[str]]


class SupportedFormats(str, enum.Enum):
    DXF = "dxf"
    DXFStream = "dxfstream"
    CSV = "csv"
    JSON = "json"
    XLXS = "xlsx"


class GenericReaderFilesHelper(BaseModel):
    file_or_buffer: Union[str, bytes, pathlib.Path, dict]
    usecols: Optional[Union[List[str], List[int]]] = None
    col_names: Optional[List[Union[str, int]]] = None
    drop_cols: Optional[List[str]] = None
    format: Optional[SupportedFormats] = None
    separator: Optional[str] = None
    index_map: Optional[dict] = None  # Adjusted for serialization
    columns_map: Optional[dict] = None  # Adjusted for serialization
    additional_reader_kwargs: dict = Field(default_factory=dict)
    encoding: str = "utf-8"
    index_col: Optional[Union[int, str, bool]] = False
    header: Union[None, int, List[int]] = 0

    # Computed fields
    file_or_buffer_type: str = Field(init=False)

    @model_validator(mode="before")
    def set_format_and_file_type(cls, values):
        file_or_buffer = values.get('file_or_buffer')
        format = values.get('format')

        # Determine format if not provided
        if format is None and file_or_buffer is not None:
            extension = get_extension(file_or_buffer)
            format_map = {
                    ".dxf" : SupportedFormats.DXF,
                    ".csv" : SupportedFormats.CSV,
                    ".json": SupportedFormats.JSON,
                    ".xlsx": SupportedFormats.XLXS,
            }
            format = format_map.get(extension.lower())
            values['format'] = format

        # Set file_or_buffer_type as a string representation
        if file_or_buffer is not None:
            values['file_or_buffer_type'] = type(file_or_buffer).__name__
        else:
            values['file_or_buffer_type'] = None

        return values

        # Custom validation for index_col to explicitly handle None

    @model_validator(mode="before")
    def validate_additional_reader_kwargs(cls, values):
        additional_reader_kwargs = values.get('additional_reader_kwargs')
        # Make sure that if any of the values is a regex expression that it is properly parsed like "delimiter":"\\\\s{2,}" to delimiter="\s{2,}"
        if additional_reader_kwargs is not None:
            for key, value in additional_reader_kwargs.items():
                if isinstance(value, str):
                    additional_reader_kwargs[key] = value.replace("\\\\", "\\")
        
        return values
    
    
    @model_validator(mode="before")
    def validate_index_col(cls, values):
        index_col = values.get('index_col')
        # Allow None explicitly
        if index_col is None:
            values['index_col'] = False
        else:
            # Ensure index_col is either int, str, or bool
            if not isinstance(index_col, (int, str, bool)):
                raise ValueError(f"Invalid value for index_col: {index_col}. Must be int, str, bool, or None.")

        return values

    # Validator to handle negative header values. If -1 is the same as null, other raise an error
    @model_validator(mode="before")
    def validate_header(cls, values):
        header = values.get('header')
        if header == -1:
            values['header'] = None
            header = None
        if header is not None and header < 0:
            raise ValueError(f"Invalid value for header: {header}. Must be None, 0, or positive integer.")
        return values
    
    # If col names is null or empy list it should be None
    @model_validator(mode="before")
    def validate_col_names(cls, values):
        col_names = values.get('col_names')
        if col_names is None or col_names == []:
            values['col_names'] = None
        return values

    @property
    def pandas_reader_kwargs(self):
        attr_dict = {
                "names"    : self.col_names,
                "header"   : self.header,
                "index_col": self.index_col,
                "usecols"  : self.usecols,
                "encoding" : self.encoding
        }
        # Check if delimiter or separator is in additional_reader_kwargs if not add it here
        if self.additional_reader_kwargs:
            delimiter = self.additional_reader_kwargs.get("delimiter", None)
        else:
            delimiter = None
        if self.separator is not None and delimiter is None:
            attr_dict["sep"] = self.separator
        
        return {**attr_dict, **self.additional_reader_kwargs}
   


import pandas as pd

from subsurface import optional_requirements


class FaultSticks:
    def __init__(self, df: pd.DataFrame):
        self.df = df

        self.pointcloud = None
        self.sticks = None

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return getattr(self, attr)
        return getattr(self.df, attr)

    def __getitem__(self, item):
        return self.df[item]

    def plot(self, notebook=False, color="black"):
        if not self.pointcloud:
            self._make_pointcloud()
            self._make_sticks()

        pv = optional_requirements.require_pyvista()   
        p = pv.Plotter(notebook=notebook)
        p.add_mesh(self.pointcloud, color=color)
        for stick in self.sticks:
            p.add_mesh(stick, color=color)
        p.show()

    def _make_pointcloud(self):
        pv = optional_requirements.require_pyvista()
        self.pointcloud = pv.PolyData(self.df[["X", "Y", "Z"]].values)

    def _make_sticks(self):
        pv = optional_requirements.require_pyvista()
        lines = []
        for stick, indices in self.df.groupby("stick id").groups.items():
            stickdf = self.df.loc[indices]
            for (r1, row1), (r2, row2) in zip(stickdf[:-1].iterrows(), stickdf[1:].iterrows()):
                line = pv.Line(
                    pointa=(row1.X, row1.Y, row1.Z),
                    pointb=(row2.X, row2.Y, row2.Z),
                )
                lines.append(line)
        self.sticks = lines


from .interfaces.stream import (
    DXF_stream_to_unstruc,
    OMF_stream_to_unstruc,
    CSV_wells_stream_to_unstruc,
    CSV_mesh_stream_to_unstruc,
    CSV_volume_stream_to_unstruc,
    CSV_volume_stream_to_struct,
    VTK_stream_to_struct,  
    MX_stream_to_unstruc,
    OBJ_stream_to_trisurf,
    GLTF_stream_to_trisurf,
    MSH_stream_to_struct,
)



from dataclasses import dataclass

from subsurface.core.reader_helpers.readers_data import GenericReaderFilesHelper


@dataclass
class ReaderUnstructuredHelper:
    reader_vertex_args: GenericReaderFilesHelper
    reader_cells_args: GenericReaderFilesHelper = None
    reader_vertex_attr_args: GenericReaderFilesHelper = None
    reader_cells_attr_args: GenericReaderFilesHelper = None



from dataclasses import dataclass

from typing import Union, List

import numpy as np
import pandas as pd

import subsurface
from subsurface import optional_requirements
from subsurface.core.structs.unstructured_elements import TriSurf
from ...visualization import to_pyvista_line, to_pyvista_mesh


def create_vertical_mesh(coords: np.ndarray, zmin: float, zmax: float):
    """
    Create a 'vertical curtain' triangular mesh by extruding the line of points
    in coords from zmin to zmax.

    Parameters
    ----------
    coords : np.ndarray
        An Nx2 array of (x, y) points along the profile.
    zmin : float
        The lower z-value (bottom of the curtain).
    zmax : float
        The upper z-value (top of the curtain).

    Returns
    -------
    vertices : np.ndarray
        An array of shape (2N, 3) with xyz-coordinates for all vertices.
    faces : np.ndarray
        An array of shape (2*(N-1), 3) with triangle indices into `vertices`.
        Each row is one triangle [vi, vj, vk].
    """
    n = len(coords)
    if n < 2:
        raise ValueError("Need at least 2 points to form a mesh.")

    # Create 2 * n vertices
    # “Bottom ring”: z = zmin
    # “Top ring”: z = zmax
    vertices = np.zeros((2 * n, 3), dtype=float)
    vertices[:n, 0:2] = coords
    vertices[:n, 2] = zmin
    vertices[n:, 0:2] = coords
    vertices[n:, 2] = zmax

    # Build faces (2 triangles per segment along the line)
    # Bottom ring indices: 0..(n-1)
    # Top ring indices: n..(2n-1)
    faces = []
    for i in range(n - 1):
        # Triangle 1: (bottom i, bottom i+1, top i)
        faces.append([i, i + 1, i + n])
        # Triangle 2: (top i, bottom i+1, top i+1)
        faces.append([i + n, i + 1, i + 1 + n])

    faces = np.array(faces, dtype=np.int64)
    return vertices, faces

@dataclass
class TexturedMesh:
    unstruct: subsurface.UnstructuredData
    struct: subsurface.StructuredData
    origin_vector3: np.array
    point_u_vector3: np.array
    point_v_vector3: np.array


def create_mesh_from_trace(linestring, zmax: Union[float, int], zmin: Union[float, int]):
    from scipy.spatial.qhull import Delaunay
    if isinstance(linestring, list):
        linestring_coords = linestring
    else:  # ! For now I leave it as this because I am not sure with object geopandas returns
        linestring_coords = linestring.coords  # * Geopandas branch
    n = len(list(linestring_coords))
    coords = np.array([[x[0] for x in list(linestring_coords)],
                       [y[1] for y in list(linestring_coords)]]).T
    # duplicating the line, once with z=lower and another with z=upper values
    vertices = np.zeros((2 * n, 3))
    vertices[:n, :2] = coords
    vertices[:n, 2] = zmin
    vertices[n:, :2] = coords
    vertices[n:, 2] = zmax
    # i+n --- i+n+1
    # |\      |
    # | \     |
    # |  \    |
    # |   \   |
    # i  --- i+1

    tri = Delaunay(vertices[:, [0, 2]])
    faces = tri.simplices
    return vertices, faces


def create_tri_surf_from_traces_texture(
        path_to_trace,
        path_to_texture: Union[List[str]],
        idx=None,
        uv=None
):
    tri_surf: list[subsurface.TriSurf] = _traces_texture_to_sub_structs(
        path_to_trace=path_to_trace,
        path_to_texture=path_to_texture,
        idx=idx,
        uv=uv
    )

    pyvista_mesh = [to_pyvista_mesh(i) for i in tri_surf]

    return tri_surf, pyvista_mesh


def lineset_from_trace(path_to_trace, idx=None):
    import geopandas as gpd

    traces = gpd.read_file(path_to_trace)
    traces = _select_traces_by_index(idx, traces)

    mesh_list = []
    for index, row in traces.iterrows():
        s = len(row['geometry'].coords.xy[0])
        vertex = np.array((*row['geometry'].coords.xy,
                           np.zeros(s))).T
        unstruct = subsurface.UnstructuredData.from_array(vertex, 'lines')
        ls = subsurface.LineSet(unstruct)
        mesh_list.append(to_pyvista_line(ls, radius=10))

    return mesh_list


def _traces_texture_to_sub_structs(path_to_trace, path_to_texture, idx, uv=None) -> list[TriSurf]:
    gpd = optional_requirements.require_geopandas()
    imageio = optional_requirements.require_imageio()

    traces = gpd.read_file(path_to_trace)
    traces = _select_traces_by_index(idx, traces)

    textured_mesh: list[TriSurf] = []
    n = 0
    for index, row in traces.iterrows():
        v, e = create_mesh_from_trace(row['geometry'], row['zmax'], row['zmin'])
        unstruct_mesh = subsurface.UnstructuredData.from_array(
            vertex=v,
            cells=e,
            vertex_attr=pd.DataFrame(np.zeros((v.shape[0], 2)), columns=['u', 'v'])
        )
        structured_data_texture = subsurface.StructuredData.from_numpy(
            array=np.array(imageio.v3.imread(path_to_texture[index]))
        )

        tri_surf = TriSurf(
            mesh=unstruct_mesh,
            texture=structured_data_texture,
            texture_origin=np.array([row['geometry'].xy[0][0], row['geometry'].xy[1][0], row['zmin']]),
            texture_point_u=np.array([row['geometry'].xy[0][-1], row['geometry'].xy[1][-1], row['zmin']]),
            texture_point_v=np.array([row['geometry'].xy[0][0], row['geometry'].xy[1][0], row['zmax']])
        )

        # * Set uv
        if uv is not None:
            uv_item = pd.DataFrame(uv[n], columns=['u', 'v'])
        else:
            uv_item = get_uv_from_pyvista(tri_surf)
        tri_surf.mesh.points_attributes = uv_item
        
        textured_mesh.append(tri_surf)
        n += 1
    
    return textured_mesh


def get_uv_from_pyvista(tri_surf: TriSurf) -> pd.DataFrame:
    pv = optional_requirements.require_pyvista()
    _mesh = to_pyvista_mesh(tri_surf)
    if tri_surf.texture is None:
        raise ValueError('unstructured_element needs texture data to be mapped.')
    _mesh.texture_map_to_plane(
        inplace=True,
        origin=tri_surf.texture_origin,
        point_u=tri_surf.texture_point_u,
        point_v=tri_surf.texture_point_v
    )
    tex = pv.numpy_to_texture(tri_surf.texture.values)
    _mesh._textures = {0: tex}
    from vtkmodules.util.numpy_support import vtk_to_numpy
    uv = vtk_to_numpy(_mesh.GetPointData().GetTCoords())

    return pd.DataFrame(uv, columns=['u', 'v'])


def _select_traces_by_index(idx, traces):
    if idx is not None:
        traces = traces.loc[idx]
    return traces



import io
from io import BytesIO
from typing import TextIO, Optional

import pandas

from ...core.structs import TriSurf
from ...core.reader_helpers.reader_unstruct import ReaderUnstructuredHelper
from ...core.reader_helpers.readers_data import GenericReaderFilesHelper
from ...core.geological_formats import BoreholeSet
from ...core.structs.base_structures import UnstructuredData, StructuredData

from ...modules import reader
from ...modules.reader.mesh._trimesh_reader import TriMeshTransformations
from ...modules.reader.volume.read_volume import read_volumetric_mesh_to_subsurface, read_VTK_structured_grid
from ...modules.reader.mesh.surfaces_api import read_2d_mesh_to_unstruct
from ...modules.reader.volume.volume_utils import interpolate_unstructured_data_to_structured_data

from ..reader.read_wells import read_wells


def DXF_stream_to_unstruc(stream: TextIO) -> UnstructuredData:
    vertex, cells, cell_attr_int, cell_attr_map = reader.dxf_stream_to_unstruct_input(stream)

    unstruct = UnstructuredData.from_array(
        vertex,
        cells,
        cells_attr=pandas.DataFrame(cell_attr_int, columns=["Id"]),
        xarray_attributes={"cell_attr_map": cell_attr_map},
    )

    return unstruct


def OMF_stream_to_unstruc(stream: BytesIO) -> list[UnstructuredData]:
    list_unstruct: list[UnstructuredData] = reader.omf_stream_to_unstructs(stream)
    return list_unstruct


def MX_stream_to_unstruc(stream: TextIO) -> list[UnstructuredData]:
    list_unstruct: list[UnstructuredData] = [reader.mx_to_unstruc_from_binary(stream)]
    return list_unstruct


def OBJ_stream_to_trisurf(obj_stream: TextIO, mtl_stream: list[TextIO],
                          texture_stream: list[io.BytesIO], coordinate_system: TriMeshTransformations) -> TriSurf:
    tri_mesh: TriSurf = reader.load_obj_with_trimesh_from_binary(
        obj_stream=obj_stream,
        mtl_stream=mtl_stream,
        texture_stream=texture_stream,
        coord_system=coordinate_system
    )
    return tri_mesh


def GLTF_stream_to_trisurf(gltf_stream: io.BytesIO, coordinate_system: TriMeshTransformations) -> TriSurf:
    tri_mesh: TriSurf = reader.load_gltf_with_trimesh(gltf_stream, coordinate_system)
    return tri_mesh


def VTK_stream_to_struct(stream: BytesIO, attribute_name: str) -> list[StructuredData]:
    struct = read_VTK_structured_grid(stream, attribute_name)
    return [struct]


def MSH_stream_to_struct(grid_stream: TextIO, values_stream: TextIO, missing_value: Optional[float], attr_name: Optional[str]) -> list[StructuredData]:
    struct = reader.read_msh_structured_grid(grid_stream, values_stream, missing_value, attr_name)
    return [struct]


def CSV_wells_stream_to_unstruc(
        collars_reader: GenericReaderFilesHelper,
        surveys_reader: GenericReaderFilesHelper,
        attrs_reader: GenericReaderFilesHelper,
        is_lith_attr: bool,
        number_nodes: int,
        add_attrs_as_nodes: bool,
        duplicate_attr_depths: bool
) -> list[UnstructuredData]:
    borehole_set: BoreholeSet = read_wells(
        collars_reader=collars_reader,
        surveys_reader=surveys_reader,
        attrs_reader=attrs_reader,
        is_lith_attr=is_lith_attr,
        number_nodes=number_nodes,
        add_attrs_as_nodes=add_attrs_as_nodes,
        duplicate_attr_depths=duplicate_attr_depths
    )

    list_unstruct: list[UnstructuredData] = [
            borehole_set.collars.data,
            borehole_set.combined_trajectory.data
    ]
    return list_unstruct


def CSV_mesh_stream_to_unstruc(
        vertex_reader: GenericReaderFilesHelper,
        edges_reader: GenericReaderFilesHelper,
        cells_attrs_reader: GenericReaderFilesHelper,
        vertex_attrs_reader: GenericReaderFilesHelper
) -> list[UnstructuredData]:
    reader_unstruc = ReaderUnstructuredHelper(vertex_reader, edges_reader, vertex_attrs_reader, cells_attrs_reader)
    ud = read_2d_mesh_to_unstruct(reader_unstruc)
    return [ud]


def CSV_volume_stream_to_unstruc(
        coord_reader: GenericReaderFilesHelper,
        attrs_reader: GenericReaderFilesHelper
) -> list[UnstructuredData]:
    ud = read_volumetric_mesh_to_subsurface(
        reader_helper_coord=coord_reader,
        reader_helper_attr=attrs_reader
    )
    return [ud]


def CSV_volume_stream_to_struct(
        coord_reader: GenericReaderFilesHelper,
        attrs_reader: GenericReaderFilesHelper,
        attr_name: str,
        resolution: list[int]
) -> list[StructuredData]:
    ud = read_volumetric_mesh_to_subsurface(
        reader_helper_coord=coord_reader,
        reader_helper_attr=attrs_reader
    )

    sd: StructuredData = interpolate_unstructured_data_to_structured_data(
        ud=ud,
        attr_name=attr_name,
        resolution=resolution
    )

    return [sd]



import numpy as np
import pandas as pd
import xarray as xr
from typing import Tuple, Optional, Union, List, Any

from subsurface import optional_requirements
from ...structs.base_structures import UnstructuredData
from ...structs.base_structures._unstructured_data_constructor import raw_attributes_to_dict_data_arrays
from ...structs.unstructured_elements import LineSet


def combine_survey_and_attrs(attrs: pd.DataFrame, survey_trajectory: LineSet,well_id_mapper: dict[str, int]) -> UnstructuredData:
    # Import moved to top for clarity and possibly avoiding repeated imports if called multiple times

    # Ensure all columns in lith exist in new_attrs, if not, add them as NaN
    new_attrs = _map_attrs_to_measured_depths(attrs, survey_trajectory, well_id_mapper)

    # Construct the final xarray dict without intermediate variable
    points_attributes_xarray_dict: dict[str, xr.DataArray] = raw_attributes_to_dict_data_arrays(
        default_attributes_name="vertex_attrs",
        n_items=survey_trajectory.data.data["vertex_attrs"].shape[0],  # TODO: Can I look this on new_attrs to remove line 11?
        dims=["points", "vertex_attr"],
        raw_attributes=new_attrs
    )

    # Inline construction of UnstructuredData
    return UnstructuredData.from_data_arrays_dict(
        xarray_dict={
                "vertex"      : survey_trajectory.data.data["vertex"],
                "cells"       : survey_trajectory.data.data["cells"],
                "vertex_attrs": points_attributes_xarray_dict["vertex_attrs"],
                "cell_attrs"  : survey_trajectory.data.data["cell_attrs"]
        },
        xarray_attributes=survey_trajectory.data.data.attrs,
        default_cells_attributes_name=survey_trajectory.data.cells_attr_name,
        default_points_attributes_name=survey_trajectory.data.vertex_attr_name
    )


def _prepare_categorical_data(attrs: pd.DataFrame) -> pd.DataFrame:
    """
    Prepare categorical data for interpolation by converting categorical columns to numeric IDs.

    Args:
        attrs: DataFrame containing attribute data

    Returns:
        Modified DataFrame with categorical data prepared for interpolation
    """
    # Create a copy to avoid modifying the original
    attrs_copy = attrs.copy()

    # If component lith exists but lith_ids doesn't, create lith_ids
    if 'component lith' in attrs_copy.columns and 'lith_ids' not in attrs_copy.columns:
        attrs_copy['lith_ids'], _ = pd.factorize(attrs_copy['component lith'], use_na_sentinel=True)

    return attrs_copy


def _prepare_new_attributes(attrs: pd.DataFrame, survey_trajectory: LineSet) -> pd.DataFrame:
    """
    Prepare the new attributes DataFrame by adding missing columns from attrs.

    Args:
        attrs: DataFrame containing attribute data
        survey_trajectory: LineSet containing trajectory data

    Returns:
        New attributes DataFrame with all necessary columns
    """
    # Start with a copy of the existing attributes DataFrame
    new_attrs = survey_trajectory.data.points_attributes.copy()

    # Add missing columns from attrs, preserving their dtypes
    for col in attrs.columns.difference(new_attrs.columns):
        new_attrs[col] = np.nan if pd.api.types.is_numeric_dtype(attrs[col]) else None

    return new_attrs


def _get_interpolation_locations(attrs_well: pd.DataFrame, well_name: str) -> np.ndarray:
    """
    Determine the locations to use for interpolation based on top and base values.

    Args:
        attrs_well: DataFrame containing well attribute data
        well_name: Name of the current well

    Returns:
        Array of location values to use for interpolation
    """
    if "base" not in attrs_well.columns:
        raise ValueError(f"Base column must be present in the file for well '{well_name}'.")
    elif "top" not in attrs_well.columns:
        return attrs_well['base'].values
    else:
        return ((attrs_well['top'] + attrs_well['base']) / 2).values


def _nearest_neighbor_categorical_interpolation(
    x_locations: np.ndarray,
    y_values: np.ndarray,
    target_depths: np.ndarray
) -> np.ndarray:
    """
    Custom nearest neighbor interpolation for categorical data.

    This function finds the nearest source point for each target point
    and assigns the corresponding categorical value.

    Args:
        x_locations: Array of source locations
        y_values: Array of categorical values at source locations
        target_depths: Array of target depths for interpolation

    Returns:
        Array of interpolated categorical values
    """
    # Initialize output array with NaN or None values
    result = np.full(target_depths.shape, np.nan, dtype=object)

    # For each target depth, find the nearest source location
    for i, depth in enumerate(target_depths):
        # Calculate distances to all source locations
        distances = np.abs(x_locations - depth)

        # Find the index of the minimum distance
        if len(distances) > 0:
            nearest_idx = np.argmin(distances)
            result[i] = y_values[nearest_idx]

    return result


def _interpolate_attribute(
    attr_values: pd.Series, 
    x_locations: np.ndarray, 
    target_depths: np.ndarray, 
    column_name: str,
    is_categorical: bool
) -> np.ndarray:
    """
    Interpolate attribute values to target depths.

    Args:
        attr_values: Series containing attribute values
        x_locations: Array of source locations for interpolation
        target_depths: Array of target depths for interpolation
        column_name: Name of the column being interpolated
        is_categorical: Whether the attribute is categorical

    Returns:
        Array of interpolated values
    """
    # For categorical data or specific columns, use custom nearest neighbor interpolation
    if is_categorical or column_name in ['lith_ids', 'component lith']:
        return _nearest_neighbor_categorical_interpolation(
            x_locations=x_locations,
            y_values=attr_values.values,
            target_depths=target_depths
        )
    else:
        # For numerical data, use scipy's interp1d with linear interpolation
        scipy = optional_requirements.require_scipy()
        interp_func = scipy.interpolate.interp1d(
            x=x_locations,
            y=attr_values.values,
            bounds_error=False,
            fill_value=np.nan,
            kind='linear'
        )
        return interp_func(target_depths)


def _map_attrs_to_measured_depths(attrs: pd.DataFrame, survey_trajectory: LineSet, well_id_mapper: dict[str, int]) -> pd.DataFrame:
    """
    Map attributes to measured depths for each well.

    Args:
        attrs: DataFrame containing attribute data
        survey_trajectory: LineSet containing trajectory data
        well_id_mapper: Dictionary mapping well names to IDs

    Returns:
        DataFrame with attributes mapped to measured depths
    """
    # Extract trajectory data
    trajectory: xr.DataArray = survey_trajectory.data.data["vertex_attrs"]
    trajectory_well_id: xr.DataArray = trajectory.sel({'vertex_attr': 'well_id'})
    measured_depths: np.ndarray = trajectory.sel({'vertex_attr': 'measured_depths'}).values.astype(np.float64)

    # Prepare data
    attrs: pd.DataFrame = _prepare_categorical_data(attrs)
    new_attrs: pd.DataFrame = _prepare_new_attributes(attrs, survey_trajectory)

    # Process each well
    for well_name in well_id_mapper:
        # Skip wells not in the attributes DataFrame
        if well_name not in attrs.index:
            print(f"Well '{well_name}' does not exist in the attributes DataFrame.")
            continue

        # Get well data
        attrs_well = attrs.loc[[well_name]]
        well_id = well_id_mapper.get(well_name)
        well_mask = (trajectory_well_id == well_id).values
        well_depths = measured_depths[well_mask]

        # Get interpolation locations
        interp_locations = _get_interpolation_locations(attrs_well, well_name)

        # Interpolate each attribute
        for col in attrs_well.columns:
            # Skip location and ID columns
            if col in ['top', 'base', 'well_id']:
                continue

            attr_values = attrs_well[col]
            is_categorical = attr_values.dtype == 'O' or isinstance(attr_values.dtype, pd.CategoricalDtype) or isinstance(attr_values.dtype, pd.StringDtype)

            # Skip columns that can't be interpolated and aren't categorical
            if is_categorical and col not in ['lith_ids', 'component lith']:
                continue

            # Interpolate and assign values
            interpolated_values = _interpolate_attribute(
                attr_values, 
                interp_locations, 
                well_depths, 
                col,
                is_categorical
            )

            # Convert to appropriate dtype to avoid pandas 3.0 dtype coercion errors
            target_dtype = new_attrs[col].dtype
            if pd.api.types.is_numeric_dtype(target_dtype) and not is_categorical:
                interpolated_values = np.asarray(interpolated_values, dtype=target_dtype)
                
            new_attrs.loc[well_mask, col] = interpolated_values

    return new_attrs



import numpy as np
import pyvista as pv


def transform_gaussian_blur(grid, sigma=20.0):
    """
    Applies a Gaussian blur to the 'model_name' field of the structured grid.

    Parameters:
        grid - PyVista grid with 'model_name' field
        sigma - Standard deviation for the Gaussian kernel
    """
    from scipy.ndimage import gaussian_filter

    # Get the original dimensions of the grid
    dims = grid.dimensions

    # Reshape the data to 3D array matching grid dimensions
    values = np.array(grid['model_name'])
    values_3d = values.reshape(dims[2] - 1, dims[1] - 1, dims[0] - 1).transpose(2, 1, 0)

    # Apply Gaussian filter
    blurred_values = gaussian_filter(values_3d, sigma=sigma, axes=(2,))

    # Reshape back to 1D array
    grid['model_name'] = blurred_values.transpose(2, 1, 0).flatten()
    return grid


def transform_sinusoidal(values, amplitude=1.0, frequency=0.01, phase=0):
    """
    Apply a sinusoidal transformation to the values.
    """
    return values + amplitude * np.sin(frequency * values + phase)


def obfuscate_model_name(grid, transform_functions, attr):
    """
    Applies transformation functions to the 'model_name' field.
    Functions can operate on either the grid or the values array.
    """
    for func in transform_functions:
        if 'grid' in func.__code__.co_varnames:
            # Function expects the full grid
            grid = func(grid)
        else:
            # Function expects just the values array
            values = np.array(grid[attr])
            grid[attr] = func(values)

    return grid


# pyvista_struct = transform_xy_to_z_propagation(pyvista_struct, z_factor=0.3, noise_level=0.1)
def transform_subtract_mean(values):
    """
    Subtract the mean of the array from each element.
    """
    return values - np.mean(values)


def transform_scale(values, scale_factor=0.003):
    """
    Multiply each value by scale_factor.
    """
    return values * scale_factor




def update_extent(pyvista_grid, new_extent):
    # new_extent: array-like with 6 elements [xmin, xmax, ymin, ymax, zmin, zmax]
    old_bounds = np.array(pyvista_grid.bounds)  # [xmin, xmax, ymin, ymax, zmin, zmax]

    # Check for valid extents
    if any(new_extent[i] >= new_extent[i + 1] for i in range(0, 6, 2)):
        raise ValueError("Each min value must be less than the corresponding max value in the new extent.")

    # Compute old ranges and new ranges for each axis
    old_ranges = old_bounds[1::2] - old_bounds[0::2]  # [x_range, y_range, z_range]
    new_ranges = np.array([new_extent[1] - new_extent[0],
                           new_extent[3] - new_extent[2],
                           new_extent[5] - new_extent[4]])

    # Avoid division by zero if any old range is zero
    if np.any(old_ranges == 0):
        raise ValueError("One of the dimensions in the current grid has zero length.")

    # Get the old points and reshape for easier manipulation
    old_points = pyvista_grid.points  # shape (N, 3)

    # Compute normalized coordinates within the old extent
    norm_points = (old_points - old_bounds[0::2]) / old_ranges

    # Compute new points based on new extent
    new_mins = np.array([new_extent[0], new_extent[2], new_extent[4]])
    new_points = new_mins + norm_points * new_ranges

    # Update the grid's points
    pyvista_grid.points = new_points

    # Updating bounds is implicit once the points are modified.
    pyvista_grid.Modified()
    return pyvista_grid



# -*- coding: utf 8 -*-
"""
Python installation file.
"""
import os
import sys
from os import path
from setuptools import setup, find_packages

if not sys.version_info[:2] >= (3, 8):
    sys.exit(f"subsurface is only meant for Python 3.8 and up.\n"
             f"Current version: {sys.version_info[0]}.{sys.version_info[1]}.")

this_directory = path.abspath(path.dirname(__file__))

readme = path.join(this_directory, "README.rst")
with open(readme, "r", encoding="utf-8") as f:
    long_description = f.read()
long_description = long_description.split('inclusion-marker')[-1]

CLASSIFIERS = [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
]


def read_requirements(file_name, base_path=""):
    # Construct the full path to the requirements file
    full_path = os.path.join(base_path, file_name)
    requirements = []
    with open(full_path, "r", encoding="utf-8") as f:
        for line in f:
            # Strip whitespace and ignore comments
            line = line.strip()
            if line.startswith("#") or not line:
                continue

            # Handle -r directive
            if line.startswith("-r "):
                referenced_file = line.split()[1]  # Extract the file name
                # Recursively read the referenced file, making sure to include the base path
                requirements.extend(read_requirements(referenced_file, base_path=base_path))
            else:
                requirements.append(line)

    return requirements


setup(
    name="subsurface_terra",
    packages=find_packages(exclude=("tests", "docs", "examples")),
    description="Subsurface data types and utilities. This version is the one used by Terranigma Solutions. Please feel free to "
                "take anything in this repository for the original one.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://softwareunderground.github.io/subsurface",
    author="Software Underground",
    author_email="hello@softwareunderground.org",
    license="Apache-2.0",
    install_requires=read_requirements("requirements.txt", "requirements"),
    extras_require={
            "plog": read_requirements("requirements_plot.txt", "requirements"),
            "opt": read_requirements("requirements_opt.txt", "requirements"),
            "dev": read_requirements("requirements_dev.txt", "requirements"),
            "all": read_requirements("requirements_all.txt", "requirements")
    },
    classifiers=CLASSIFIERS,
    zip_safe=False,
    use_scm_version={
            "root"       : ".",
            "relative_to": __file__,
            "write_to"   : path.join("subsurface", "_version.py"),
    },
    setup_requires=["setuptools_scm"],
)



from dataclasses import dataclass
from typing import Union, Hashable, Optional

import pandas as pd

from ._map_attrs_to_survey import combine_survey_and_attrs
from ._survey_to_unstruct import data_frame_to_unstructured_data
from ...structs.base_structures import UnstructuredData
from ...structs.unstructured_elements import LineSet

NUMBER_NODES = 30
RADIUS = 10


@dataclass
class Survey:
    ids: list[str]
    survey_trajectory: LineSet
    well_id_mapper: dict[str, int] = None  #: This is following the order of the survey csv that can be different that the collars

    @property
    def id_to_well_id(self):
        # Reverse the well_id_mapper dictionary to map IDs to well names
        id_to_well_name_mapper = {v: k for k, v in self.well_id_mapper.items()}
        return id_to_well_name_mapper

    @classmethod
    def from_df(cls, survey_df: 'pd.DataFrame', attr_df: Optional['pd.DataFrame'] = None, number_nodes: Optional[int] = NUMBER_NODES,
                duplicate_attr_depths: bool = False) -> 'Survey':
        """
        Create a Survey object from two DataFrames containing survey and attribute data.

        :param survey_df: DataFrame containing survey data.
        :param attr_df: DataFrame containing attribute data. This is used to make sure the raw data is perfectly aligned.
        :param number_nodes: Optional parameter specifying the number of nodes.
        :return: A Survey object representing the input data.

        """
        trajectories: UnstructuredData = data_frame_to_unstructured_data(
            survey_df=_correct_angles(survey_df),
            attr_df=attr_df,
            number_nodes=number_nodes,
            duplicate_attr_depths=duplicate_attr_depths
        )
        # Grab the unique ids
        unique_ids = trajectories.points_attributes["well_id"].unique()

        return cls(
            ids=unique_ids,
            survey_trajectory=LineSet(data=trajectories, radius=RADIUS),
            well_id_mapper=trajectories.data.attrs["well_id_mapper"]
        )

    def get_well_string_id(self, well_id: int) -> str:
        return self.ids[well_id]

    def get_well_num_id(self, well_string_id: Union[str, Hashable]) -> int:
        return self.well_id_mapper.get(well_string_id, None)

    def update_survey_with_lith(self, lith: pd.DataFrame):
        unstruct: UnstructuredData = combine_survey_and_attrs(lith, self.survey_trajectory, self.well_id_mapper)
        self.survey_trajectory.data = unstruct

    def update_survey_with_attr(self, attrs: pd.DataFrame):
        self.survey_trajectory.data = combine_survey_and_attrs(attrs, self.survey_trajectory, self.well_id_mapper)


def _correct_angles(df: pd.DataFrame) -> pd.DataFrame:
    def correct_inclination(inc: float) -> float:
        if inc < 0:
            inc = inc % 360  # Normalize to 0-360 range first if negative
        if 0 <= inc <= 180:
            # add or subtract a very small number to make sure that 0 or 180 are never possible
            return inc + 1e-10 if inc == 0 else inc - 1e-10
        elif 180 < inc < 360:
            return 360 - inc  # Reflect angles greater than 180 back into the 0-180 range
        else:
            raise ValueError(f'Inclination value {inc} is out of the expected range of 0 to 360 degrees')

    def correct_azimuth(azi: float) -> float:
        return azi % 360  # Normalize azimuth to 0-360 range

    df['inc'] = df['inc'].apply(correct_inclination)
    df['azi'] = df['azi'].apply(correct_azimuth)

    return df



def require_omf():
    try:
        import omfvista
    except ImportError:
        raise ImportError("The omfvista package is required to run this function.")
    return omfvista


def require_rasterio():
    try:
        import rasterio
    except ImportError:
        raise ImportError("The rasterio package is required to run this function.")
    return rasterio


def require_pyvista():
    try:
        import pyvista
    except ImportError:
        raise ImportError("The pyvista package is required to run this function.")
    return pyvista


def require_imageio():
    try:
        import imageio
    except ImportError:
        raise ImportError("The imageio package is required to run this function.")
    return imageio


def require_geopandas():
    try:
        import geopandas as gpd
    except ImportError:
        raise ImportError("The geopandas package is required to run this function.")
    return gpd


def require_segyio():
    try:
        import segyio
    except ImportError:
        raise ImportError("The segyio package is required to run this function.")
    return segyio


def require_welly():
    try:
        import welly 
    except ImportError:
        raise ImportError("The welly package is required to run this function.")
    return welly


def require_striplog():
    try:
        import striplog
    except ImportError:
        raise ImportError("The striplog package is required to run this function.")
    return striplog


def require_ezdxf():
    try:
        import ezdxf
    except ImportError:
        raise ImportError("The ezdxf package is required to run this function.")
    return ezdxf


def require_trimesh():
    try:
        import trimesh
    except ImportError:
        raise ImportError("The trimesh package is required to run this function.")
    return trimesh


def require_scipy():
    try:
        import scipy
    except ImportError:
        raise ImportError("The scipy package is required to run this function.")
    return scipy


def require_pandas():
    try:
        import pandas as pd
    except ImportError:
        raise ImportError("The pandas package is required to run this function.")
    return pd


def require_wellpathpy():
    try:
        import wellpathpy
    except ImportError:
        raise ImportError("The wellpathpy package is required to run this function.")
    return wellpathpy


def require_subsurface():
    try:
        import subsurface
    except ImportError:
        raise ImportError("The subsurface package is required to run this function.")
    return subsurface

def require_matplotlib():
    try:
        import matplotlib
    except ImportError:
        raise ImportError("The matplotlib package is required to run this function.")
    return matplotlib


import numpy as np
from ..base_structures import StructuredData


class StructuredGrid:
    # TODO check structured_data has three coordinates
    """Container for curvilinear mesh grids.

        This is analogous to PyVista's StructuredGrid class or discretize's
        CurviMesh class.

    """

    def __init__(self, structured_data: StructuredData):
        self.ds = structured_data

    @property
    def cartesian_dimensions(self):
        return len(self.cartesian_coords_names)

    @property
    def cartesian_coords_names(self):
        coord_names = np.array(['X', 'Y', 'Z', 'x', 'y', 'z'])
        return coord_names[np.isin(coord_names, self.ds.data.dims)]

    @property
    def coord(self):
        return self.ds.data.coords

    @property
    def meshgrid_3d(self):
        cart_coord = [self.coord[i] for i in self.cartesian_coords_names]
        grid_3d = np.meshgrid(*cart_coord, indexing='ij')
        return grid_3d

    @property
    def active_attributes(self) -> np.ndarray:
        return self.ds.data[self.ds.active_data_array_name].values

    def meshgrid_2d(self, attribute_name_coord_name: str = None) -> list:
        """

        Args:
            attribute_name_coord_name(str): Name of the xarray.Dataset coord that
             will be used for the z direction. This must be 2d

        Returns:

        """
        grid_2d_: tuple = np.meshgrid(self.coord['x'], self.coord['y'])
        grid_2d = list(grid_2d_)
        if attribute_name_coord_name is not None:
            z_coord = self.ds.data[attribute_name_coord_name].values.T
            if z_coord.ndim != 2:
                raise AttributeError('The attribute must be a 2D array')

            grid_2d.append(z_coord)

        return grid_2d



import pandas as pd


def safe_convert_to_float32(df: pd.DataFrame, error_handling: str = 'raise') -> pd.DataFrame:
    """
    Convert DataFrame columns to float32, handling non-convertible columns.

    Args:
        df: DataFrame to convert
        error_handling: How to handle non-convertible columns:
            - 'raise': Raise an error
            - 'skip': Skip non-convertible columns
            - 'drop': Drop non-convertible columns

    Returns:
        DataFrame with converted columns
    """
    convertible, non_convertible = _check_convertible_to_float32(df)

    if non_convertible:
        if error_handling == 'raise':
            raise ValueError(
                f"Cannot convert columns to float32: {non_convertible}. "
                f"These columns contain non-numeric data."
            )
        elif error_handling == 'skip':
            # Only convert the convertible columns
            result = df.copy()
            for col in convertible:
                result[col] = df[col].astype('float32')
            return result
        elif error_handling == 'drop':
            # Drop non-convertible columns
            return df[convertible].astype('float32')
        else:
            raise ValueError(f"Invalid error_handling: {error_handling}")

    return df.astype('float32')


def _check_convertible_to_float32(df: pd.DataFrame) -> tuple[list[str], list[str]]:
    """
    Check which columns in a DataFrame can be safely converted to float32.

    Returns:
        tuple: (convertible_columns, non_convertible_columns)
    """
    convertible = []
    non_convertible = []

    for col in df.columns:
        if pd.api.types.is_numeric_dtype(df[col]):
            # Already numeric, can convert
            convertible.append(col)
        elif pd.api.types.is_bool_dtype(df[col]):
            # Boolean can be converted (True->1.0, False->0.0)
            convertible.append(col)
        elif pd.api.types.is_string_dtype(df[col]) or pd.api.types.is_object_dtype(df[col]):
            # Try to convert to numeric
            try:
                pd.to_numeric(df[col], errors='raise')
                convertible.append(col)
            except (ValueError, TypeError):
                non_convertible.append(col)
        else:
            # Other types (datetime, timedelta, categorical, etc.)
            non_convertible.append(col)

    return convertible, non_convertible



from typing import Union, TextIO
import io

from ._trimesh_reader import load_with_trimesh, trimesh_to_unstruct, TriMeshReaderFromBlob, TriMeshTransformations
from ....core.structs import TriSurf



def load_obj_with_trimesh_from_binary(obj_stream: TextIO, mtl_stream: list[TextIO], 
                                      texture_stream: list[io.BytesIO], coord_system: TriMeshTransformations) -> TriSurf:
    tri_surf: TriSurf = TriMeshReaderFromBlob.OBJ_stream_to_trisurf(
        obj_stream=obj_stream,
        mtl_stream=mtl_stream,
        texture_stream=texture_stream,
        coord_system=coord_system
    )
    
    return tri_surf
    

def load_obj_with_trimesh(path_to_obj: str, plot: bool = False) -> TriSurf:
    """
    Load and process an OBJ file, returning trimesh-compatible objects.

    This function loads an OBJ file using `trimesh`, optionally plots it,
    and converts the loaded mesh or scene into a suitable unstructured format
    using the `trimesh_to_unstruct` function. Depending on the input and the 
    contents of the OBJ file, it may return a Trimesh object or a Scene object.

    Note: 
    This implementation does not include the capability to force a PNG as a 
    texture if the material does not already have an associated image. 
    `trimesh` ignores UVs when this condition occurs. Modifications to 
    `trimesh`'s loading function would be necessary to address this limitation.

    Args:
        path_to_obj: Path to the OBJ file to be loaded.
                     This must be a valid file path to a 3D object in OBJ format.
        plot: Boolean flag indicating whether to visually plot the loaded model.
              Defaults to False.

    Returns:
        A `trimesh.Trimesh` object if a single mesh is loaded, or a `trimesh.Scene`
        object if the file contains multiple meshes or a scene.

    Raises:
        `FileNotFoundError`: If the provided file path does not exist.
        `ValueError`: If the OBJ file could not be properly processed.

    """
    trimesh = load_with_trimesh(path_to_obj, file_type="obj", plot=plot)
    trisurf = trimesh_to_unstruct(trimesh)
    return trisurf



import numpy as np

__all__ = ['write_data_block_header', 'encode']

rexFileHeaderSize = 64
rexCoordSize = 22

file_header_size = 86
rexDataBlockHeaderSize = 16

file_header_and_data_header = 102
mesh_header_size = 128
all_header_size = 230

# Supported block types
# typeLineSet = 0
# typeText = 1
# typePointList = 2
typeMesh = 3
# typeImage = 4
# typeMaterial = 5
# typePeopleSimulation = 6
# typeUnityPackage = 7
# typeSceneNode = 8


n_bytes = 0


def write_data_block_header(size_data, data_id=1, data_type=3, version_data=1):
    """Function to write a DATA BLOCK header.

    Args:
        size_data: data block size (without header)
        data_id: id which is used in the database
        data_type (int): Type of data the data block contains:
            * 0	LineSet	A list of vertices which get connected by line segments
            * 1	Text	A position information and the actual text
            * 2	PointList	A list of 3D points with color information (e.g. point cloud)
            * 3	Mesh	A triangle mesh datastructure️
            * 4	Image	A single of arbitrary format can be stored in this block
            * 5	MaterialStandard	A standard (mesh) material definition
            * 6	SceneNode	A wrapper around a data block which can be used in the scenegraph
            * 7	Track	A track is a tracked position and orientation of an AR device
        version_data: version for this data block

    Returns:

    """

    input_ = [(data_type, 'uint16'),  # data type
              (version_data, 'uint16'),  # version for this data block
              (size_data, 'uint32'),  # data block size (without header)
              (data_id, 'uint64')]  # id which is used in the database

    block_bytes = encode(input_)
    return block_bytes


def encode(input_: list):
    """Encode python objects - normally Python primitives or numpy arrays - into
    its correspondent byte representation

    Args:
        input_ (List[tuples]): List of tuples: (object, type)

    Returns:
        byte: Array of bytes
    """
    global n_bytes
    block = bytearray()

    for tup in input_:
        arr = np.array(tup[0], dtype=tup[1]).tobytes()
        n_bytes += len(arr)
        block += arr

    return block
